from unittest import TestCase

import numpy as np

from zephyr_ml.primitives.postprocessing import FindThreshold


class FindThresholdTest(TestCase):

    y = np.array([1, 1, 0, 0, 1, 1, 1, 0])
    y_hat_1d = np.array([0.8, 0.9, 0.6, 0.5, 0.85, 0.7, 0.95, 0.2])
    y_hat_2d = np.array([[0.2, 0.8],
                         [0.1, 0.9],
                         [0.4, 0.6],
                         [0.5, 0.5],
                         [0.15, 0.85],
                         [0.3, 0.7],
                         [0.05, 0.95],
                         [0.8, 0.2]])

    def _run(self, y, y_hat, value):
        threshold = FindThreshold()
        threshold.fit(y, y_hat)

        assert threshold._threshold == value
        binary_y_hat, detected_threshold, scores = threshold.apply_threshold(
            y_hat)
        np.testing.assert_allclose(binary_y_hat, y)

    def test_1d(self):
        self._run(self.y, self.y_hat_1d, 0.6)

    def test_2d(self):
        self._run(self.y, self.y_hat_2d, 0.6)



from zephyr_ml.labeling.utils import denormalize


def converter_replacement_presence(es, column_map={}):
    """Calculates the converter replacement presence.

    Args:
        es (ft.EntitySet):
            EntitySet of data to check converter replacements.
        column_map (dict):
            Optional dictionary to update default column names to the
            actual corresponding column names in the data slice. Can contain the
            following keys:
                "sap_code": Column that contains the material SAP code. Defaults
                    to "COD_MATERIAL_SAP".
                "turbine_id": Column containing the ID of the turbine associated with a
                    stoppage. Must match the index column of the 'turbines' entity.
                    Defaults to "COD_ELEMENT".
                "description": Column containing the description for a given notification.
                    Defaults to "DES_MEDIUM".
                "time_index": Column to use as the time index for the data slice.
                    Defaults to "DAT_MALF_START".


    Returns:
        label:
            Labeling function to find converter replacement presence over a data slice.
        df:
            Denormalized dataframe of data to get labels from.
        meta:
            Dictionary containing metadata about labeling function.

    """
    sap_code = column_map.get('sap_code', 'COD_MATERIAL_SAP')
    column_map.get('description', 'DES_MEDIUM')
    turbine_id = column_map.get('turbine_id_column', 'COD_ELEMENT')
    time_index = column_map.get('time_index', 'DAT_MALF_START')

    def label(ds, **kwargs):
        logic1 = (ds[sap_code] == 36052411).any()
        # logic2 = ds[DESCRIPTION].str.lower().apply(lambda x: 'inu' in x).any()
        f = logic1  # or logic2
        return f

    meta = {
        "target_entity_index": turbine_id,
        "time_index": time_index,
        "window_size": "10d"
    }

    # denormalize(es, entities=['notifications', 'work_orders'])
    df = denormalize(es, entities=['notifications'])
    df = df.dropna(subset=[time_index])

    return label, df, meta



import copy
import json
import logging
import os
from functools import wraps
from inspect import getfullargspec

import composeml as cp
import featuretools as ft
import numpy as np
import pandas as pd
from mlblocks import MLBlock, MLPipeline
from sklearn.model_selection import train_test_split

from zephyr_ml.entityset import VALIDATE_DATA_FUNCTIONS, _create_entityset
from zephyr_ml.feature_engineering import process_signals
from zephyr_ml.labeling import get_labeling_functions, get_labeling_functions_map

DEFAULT_METRICS = [
    "sklearn.metrics.accuracy_score",
    "sklearn.metrics.precision_score",
    "sklearn.metrics.f1_score",
    "sklearn.metrics.recall_score",
    "zephyr_ml.primitives.postprocessing.confusion_matrix",
    "zephyr_ml.primitives.postprocessing.roc_auc_score_and_curve",
]

LOGGER = logging.getLogger(__name__)


class GuideHandler:

    def __init__(self, producers_and_getters, set_methods):
        self.cur_iteration = 0
        self.current_step = -1
        self.start_point = -1
        self.producers_and_getters = producers_and_getters
        self.set_methods = set_methods

        self.producer_to_step_map = {}
        self.getter_to_step_map = {}

        self.iterations = []
        for idx, (producers, getters) in enumerate(self.producers_and_getters):
            self.iterations.append(-1)

            for prod in producers:
                self.producer_to_step_map[prod.__name__] = idx

            for get in getters:
                self.getter_to_step_map[get.__name__] = idx

    def get_necessary_steps(self, actual_next_step):
        step_strs = []
        for step in range(self.current_step, actual_next_step):
            option_strs = []
            for opt in self.producers_and_getters[step][0]:
                option_strs.append(opt.__name__)
            step_strs.append(f"{step}. {' or '.join(option_strs)}")
        return "\n".join(step_strs)

    def get_get_steps_in_between(self, cur_step, next_step):
        step_strs = []
        for step in range(cur_step + 1, next_step):
            step_strs.append(
                f"{step} {self.producers_and_getters[step][1][0]}")
        return step_strs

    def get_last_up_to_date(self, next_step):
        latest_up_to_date = 0
        for step in range(next_step):
            if self.iterations[step] == self.cur_iteration:
                latest_up_to_date = step
        return latest_up_to_date

    def join_steps(self, step_strs):
        return "\n".join(step_strs)

    def get_steps_in_between(self, cur_step, next_step):
        step_strs = []
        for step in range(cur_step + 1, next_step):
            option_strs = []
            for opt in self.producers_and_getters[step][0]:
                option_strs.append(opt.__name__)
            step_strs.append(f"{step}. {' or '.join(option_strs)}")
        return step_strs

    def log_next_producer_step(self, name):
        next_step = self.current_step + 1

        if next_step >= len(self.producers_and_getters):
            LOGGER.warning("[GUIDE] You have reached the end of the \
                           predictive engineering workflow.\
                           You may continue to go back and reperform steps based on results.")
        else:
            next_step_name = self.producers_and_getters[next_step][0][0].__name__
            LOGGER.warning(f"[GUIDE] Successfully performed {name}. You can perform the \
                           next step by calling\
                            {next_step_name}.")

    def perform_producer_step(self, zephyr, method,
                              *method_args, **method_kwargs):
        step_num = self.producer_to_step_map[method.__name__]
        res = method(zephyr, *method_args, **method_kwargs)
        self.current_step = step_num
        self.iterations[step_num] = self.cur_iteration
        self.log_next_producer_step(method.__name__)
        return res

    def try_log_forward_set_method_warning(self, name, next_step):
        LOGGER.warning(
            f"[GUIDE] STALE WARNING:  Performing step {next_step} \
                    from step {self.current_step} \
                    via {name}.\
                    This is a forward step via a set method.\
                    The current step is {self.current_step}.\
                    All previous steps will be considered stale.")

    def try_log_backwards_set_method_warning(self, name, next_step):
        LOGGER.warning(f"[GUIDE] STALE WARNING: Performing step \
                       {next_step} from step {self.current_step} \
                       via {name}. \
                    This is a backwards step via a set method.\
                        All other steps will be considered stale.")

    def try_log_backwards_key_method_warning(self, name, next_step):
        steps_in_between = self.join_steps(
            self.get_steps_in_between(
                self.current_step, next_step))
        LOGGER.warning(f"[GUIDE] STALE WARNING: Performing step {next_step} \
                       from step {self.current_step} via {name}. \
                    This is a backwards step via a key method.\
                        The following steps will be considered stale:\n{steps_in_between}")

    def log_get_inconsistent_warning(self, name, next_step):
        prod_steps_str = ' or '.join([method.__name__ for method in
                                      self.producers_and_getters[next_step][0]])
        prod_steps = f"{next_step}.{prod_steps_str}"
        latest_up_to_date = self.get_last_up_to_date(next_step)
        LOGGER.warning(f"[GUIDE] INCONSISTENCY WARNING: Unable to perform {name} \
                       because {prod_steps} has not \
                        been run yet. Run steps starting at or before \
                        {latest_up_to_date} ")

    def log_get_stale_warning(self, name, next_step):
        latest_up_to_date = self.get_last_up_to_date(next_step)
        LOGGER.warning(f"[GUIDE] STALE WARNING: Performing {name}. \
                       This data is potentially stale. \
                        Re-run steps starting at or before \
                        {latest_up_to_date} to ensure data is up to date.")

    # tries to perform step if possible -> warns that data might be stale

    def try_perform_forward_producer_step(
            self, zephyr, method, *method_args, **method_kwargs):
        name = method.__name__
        next_step = self.producer_to_step_map[name]
        if name in self.set_methods:  # set method will update start point and start new iteration
            self.try_log_forward_set_method_warning(name, next_step)
            self.start_point = next_step
            self.cur_iteration += 1
        # next_step == 0, set method (already warned), or previous step is up
        # to term
        res = self.perform_producer_step(
            zephyr, method, *method_args, **method_kwargs)
        return res

    def try_perform_backward_producer_step(
            self, zephyr, method, *method_args, **method_kwargs):
        name = method.__name__
        next_step = self.producer_to_step_map[name]
        # starting new iteration
        self.cur_iteration += 1
        if next_step == 0 or name in self.set_methods:
            self.start_point = next_step
        else:  # key method
            # mark everything from start point to next step as current term
            for i in range(self.start_point, next_step):
                if self.iterations[i] != -1:
                    self.iterations[i] = self.cur_iteration

        if name in self.set_methods:
            self.try_log_backwards_set_method_warning(name, next_step)
        else:
            self.try_log_backwards_key_method_warning(name, next_step)

        res = self.perform_producer_step(
            zephyr, method, *method_args, **method_kwargs)

        return res

    def try_perform_producer_step(
            self, zephyr, method, *method_args, **method_kwargs):
        name = method.__name__
        next_step = self.producer_to_step_map[name]
        if next_step >= self.current_step:
            res = self.try_perform_forward_producer_step(
                zephyr, method, *method_args, **method_kwargs)
            return res
        else:
            res = self.try_perform_backward_producer_step(
                zephyr, method, *method_args, **method_kwargs)
            return res

    # dont update current step or terms

    def try_perform_inconsistent_producer_step(  # add using stale and overwriting
            self, zephyr, method, *method_args, **method_kwargs):
        name = method.__name__
        next_step = self.producer_to_step_map[name]
        # inconsistent forward step: performing key method but previous step is
        # not up to date
        if (next_step >= self.current_step and
                self.iterations[next_step - 1] != self.cur_iteration):
            corr_set_method = self.producers_and_getters[next_step][0][1].__name__
            prev_step = next_step - 1
            prev_set_method = self.producers_and_getters[prev_step][0][1].__name__
            prev_key_method = self.producers_and_getters[prev_step][0][0].__name__
            LOGGER.warning(f"[GUIDE] INCONSISTENCY WARNING:Unable \
                           to perform {name} because you are\
                            performing a key method at\
                            step {next_step} but the result of the previous step, \
                            step {prev_step}, is STALE.\
                           If you already have the data for step {next_step}, \
                            you can use the corresponding set method: {corr_set_method}.\
                            Otherwise, please perform step {prev_step} \
                                with {prev_key_method} or {prev_set_method}.")
        # inconsistent backward step: performing set method at nonzero step
        # elif next_step < self.current_step and name in self.set_method:
        #     first_set_method = self.producers_and_getters[0][0][1].__name__
        #     corr_key_method = self.producers_and_getters[next_step][0][0].__name__
        #     LOGGER.warning(f"Unable to perform {name} because you are going backwards \
        #                    and performing step {next_step} with a set method.\
        #                    You can only perform a backwards step with a set \
        #                     method at step 0: {first_set_method}.\
        #                     If you would like to perform step {next_step}, \
        #                         please use the corresponding key method: {corr_key_method}.")
        # inconsistent backward step: performing key method but previous step
        # is not up to date
        elif (next_step < self.current_step and
              self.iterations[next_step - 1] != self.cur_iteration):
            prev_step = next_step - 1
            prev_key_method = self.producers_and_getters[prev_step][0][0].__name__
            corr_set_method = self.producers_and_getters[next_step][0][1].__name__
            prev_get_method = self.producers_and_getters[prev_step][1][0].__name__
            prev_set_method = self.producers_and_getters[prev_step][0][1].__name__
            LOGGER.warning(f"[GUIDE] INCONSISTENCY WARNING: Unable to perform {name} \
                           because you are going \
                           backwards and starting a new iteration by\
                           performing a key method at step {next_step} \
                            but the result of the previous step,\
                            step {prev_step}, is STALE.\
                            If you want to use the STALE result of the PREVIOUS step, \
                            you can call {prev_get_method} to get the data, then\
                            {prev_set_method} to set the data, and then recall this method.\
                            If you want to regenerate the data of the PREVIOUS step, \
                            please call {prev_key_method}, and then recall this method.\
                            If you already have the data for THIS step, you can \
                            call {corr_set_method} to set the data.\
                            ")

    def try_perform_getter_step(
            self, zephyr, method, *method_args, **method_kwargs):
        name = method.__name__
        # either inconsistent, stale, or up to date
        step_num = self.getter_to_step_map[name]
        step_iteration = self.iterations[step_num]
        if step_iteration == -1:
            self.log_get_inconsistent_warning(name, step_num)
        elif step_iteration == self.cur_iteration:
            res = method(zephyr, *method_args, **method_kwargs)
            return res
        else:
            self.log_get_stale_warning(name, step_num)
            res = method(zephyr, *method_args, **method_kwargs)
            return res

    def guide_step(self, zephyr, method, *method_args, **method_kwargs):
        method_name = method.__name__
        if method_name in self.producer_to_step_map:
            # up-todate
            next_step = self.producer_to_step_map[method_name]
            if (next_step == 0 or  # 0 step always valid, starting new iteration
                # set method always valid, but will update start point and
                # start new iteration
                method_name in self.set_methods or
                    # key method valid if previous step is up to date
                    self.iterations[next_step - 1] == self.cur_iteration):
                # forward step only valid if set method or key method w/ no
                # skips
                res = self.try_perform_producer_step(
                    zephyr, method, *method_args, **method_kwargs)
                return res
            else:  # stale or inconsistent
                res = self.try_perform_inconsistent_producer_step(
                    zephyr, method, *method_args, **method_kwargs)
                return res
        elif method_name in self.getter_to_step_map:
            res = self.try_perform_getter_step(
                zephyr, method, *method_args, **method_kwargs)
            return res
        else:
            print(f"Method {method_name} does not need to be wrapped")


def guide(method):

    @wraps(method)
    def guided_step(instance, *method_args, **method_kwargs):
        return instance._guide_handler.guide_step(
            instance, method, *method_args, **method_kwargs)

    return guided_step


class Zephyr:
    """Zephyr Class.

    The Zephyr Class supports all the steps of the predictive engineering workflow
    for wind farm operations data. It manages user state and handles entityset creation, labeling,
    feature engineering, model training and evaluation.
    """

    def __init__(self):
        """Initialize a new Zephyr instance."""
        self._entityset = None

        self._label_times = None
        self._label_times_meta = None

        self._label_col_name = "label"
        self._feature_matrix = None

        self._pipeline = None

        self._X_train = None
        self._X_test = None
        self._y_train = None
        self._y_test = None

        # tuple of 2 arrays: producers and attributes
        step_order = [
            ([
                self.generate_entityset, self.set_entityset], [
                self.get_entityset]), ([
                    self.generate_label_times, self.set_label_times], [
                    self.get_label_times]), ([
                        self.generate_feature_matrix, self.set_feature_matrix], [
                            self.get_feature_matrix]), ([
                                self.generate_train_test_split, self.set_train_test_split], [
                                    self.get_train_test_split]), ([
                                        self.fit_pipeline, self.set_fitted_pipeline], [
                                            self.get_fitted_pipeline]), ([
                                                self.predict, self.evaluate], [])]
        set_methods = set([self.set_entityset.__name__,
                           self.set_label_times.__name__,
                           self.set_feature_matrix.__name__,
                           self.set_train_test_split.__name__,
                           self.set_fitted_pipeline.__name__])
        self._guide_handler = GuideHandler(step_order, set_methods)

    def GET_ENTITYSET_TYPES(self):
        """Get the supported entityset types and their required dataframes/columns.

        Returns:
            dict: A dictionary mapping entityset types (PI/SCADA/Vibrations) to their
                descriptions and value.
        """
        info_map = {}
        for es_type, val_fn in VALIDATE_DATA_FUNCTIONS.items():
            info_map[es_type] = {"obj": es_type,
                                 "desc": " ".join((val_fn.__doc__.split()))}

        return info_map

    def GET_LABELING_FUNCTIONS(self):
        """Get the available predefined labeling functions.

        Returns:
            dict: A dictionary mapping labeling function names to their
            descriptions and implementations.
        """
        return get_labeling_functions()

    def GET_EVALUATION_METRICS(self):
        """Get the available evaluation metrics.

        Returns:
            dict: A dictionary mapping metric names to their descriptions
             and MLBlock instances.
        """
        info_map = {}
        for metric in DEFAULT_METRICS:
            primitive = self._get_ml_primitive(metric)
            info_map[metric] = {"obj": primitive,
                                "desc": primitive.metadata["description"]}
        return info_map

    @guide
    def generate_entityset(
            self,
            dfs,
            es_type,
            custom_kwargs_mapping=None,
            signal_dataframe_name=None,
            signal_column=None,
            signal_transformations=None,
            signal_aggregations=None,
            signal_window_size=None,
            signal_replace_dataframe=False,
            **sigpro_kwargs):
        """Generate an entityset from input dataframes with optional signal processing.

        Args:
            dfs (dict): Dictionary mapping entity names to pandas DataFrames.
            es_type (str): Type of signal data, either 'SCADA' or 'PI'.
            custom_kwargs_mapping (dict, optional): Custom keyword arguments
                for entityset creation.
            signal_dataframe_name (str, optional): Name of dataframe containing
                signal data to process.
            signal_column (str, optional): Name of column containing signal values to process.
            signal_transformations (list[dict], optional): List of transformation
                primitives to apply.
            signal_aggregations (list[dict], optional): List of aggregation primitives to apply.
            signal_window_size (str, optional): Size of window for signal binning (e.g. '1h').
            signal_replace_dataframe (bool, optional): Whether to replace
                original signal dataframe.
            **sigpro_kwargs: Additional keyword arguments for signal processing.

        Returns:
            featuretools.EntitySet: EntitySet containing the processed data and relationships.
        """
        entityset = _create_entityset(dfs, es_type, custom_kwargs_mapping)

        # perform signal processing
        if signal_dataframe_name is not None and signal_column is not None:
            if signal_transformations is None:
                signal_transformations = []
            if signal_aggregations is None:
                signal_aggregations = []
            process_signals(
                entityset,
                signal_dataframe_name,
                signal_column,
                signal_transformations,
                signal_aggregations,
                signal_window_size,
                signal_replace_dataframe,
                **sigpro_kwargs)

        self._entityset = entityset
        return self._entityset

    @guide
    def set_entityset(self, entityset=None, es_type=None, entityset_path=None,
                      custom_kwargs_mapping=None):
        """Set the entityset for this Zephyr instance.

        Args:
            entityset (featuretools.EntitySet, optional): An existing entityset to use.
            es_type (str, optional): The type of entityset (pi/scada/vibrations).
            entityset_path (str, optional): Path to a saved entityset to load.
            custom_kwargs_mapping (dict, optional): Custom keyword arguments for validation.

        Raises:
            ValueError: If no entityset is provided through any of the parameters.
        """
        if entityset_path is not None:
            entityset = ft.read_entityset(entityset_path)

        if entityset is None:
            raise ValueError(
                "No entityset passed in. Please pass in an entityset object\
                via the entityset parameter or an entityset path via the \
                entityset_path parameter.")

        dfs = entityset.dataframe_dict

        validate_func = VALIDATE_DATA_FUNCTIONS[es_type]
        validate_func(dfs, custom_kwargs_mapping)

        self._entityset = entityset

    @guide
    def get_entityset(self):
        """Get the current entityset.

        Returns:
            featuretools.EntitySet: The current entityset.

        Raises:
            ValueError: If no entityset has been set.
        """
        if self._entityset is None:
            raise ValueError(
                "No entityset has been created or set in this instance.")

        return self._entityset

    @guide
    def generate_label_times(
        self, labeling_fn, num_samples=-1, subset=None, column_map={}, verbose=False, thresh=None,
        window_size=None, minimum_data=None, maximum_data=None, gap=None, drop_empty=True, **kwargs
    ):
        """Generate label times using a labeling function.

        This method applies a labeling function to the entityset to generate labels at specific
        timestamps. The labeling function can be either a predefined one (specified by name) or
        a custom callable.

        Args:
            labeling_fn (callable or str): Either a custom labeling function or the
                name of a predefined function (e.g. 'brake_pad_presence').
                Predefined functions like brake_pad_presence analyze specific patterns
                in the data (e.g. brake pad mentions in stoppage comments) and
                return a tuple containing:
                1) A label generation function that processes data slices
                2) A denormalized dataframe containing the source data
                3) Metadata about the labeling process (e.g. target entity, time index)
            num_samples (int, optional): Number of samples to generate. -1 for all. Defaults to -1.
            subset (int or float, optional): Number or fraction of samples to randomly select.
            column_map (dict, optional): Mapping of column names for the labeling function.
            verbose (bool, optional): Whether to display progress. Defaults to False.
            thresh (float, optional): Threshold for label binarization. If None, tries to
                use threshold value from labeling function metadata, if any.
            window_size (str, optional): Size of the window for label generation (e.g. '1h').
                If None, tries to use window size value from labeling function metadata, if any.
            minimum_data (str, optional): Minimum data required before cutoff time.
            maximum_data (str, optional): Maximum data required after cutoff time.
            gap (str, optional): Minimum gap between consecutive labels.
            drop_empty (bool, optional): Whether to drop windows with no events. Defaults to True.
            **kwargs: Additional arguments passed to the label generation function.

        Returns:
            tuple: (composeml.LabelTimes, dict) The generated label times and metadata.
                Label times contain the generated labels at specific timestamps.
                Metadata contains information about the labeling process.

        Raises:
            ValueError: If labeling_fn is a string but not a recognized predefined function.
            AssertionError: If entityset has not been generated or set or labeling_fn is
                not a string and not callable.
        """
        assert self._entityset is not None, "entityset has not been set"

        if isinstance(labeling_fn, str):  # get predefined labeling function
            labeling_fn_map = get_labeling_functions_map()
            if labeling_fn in labeling_fn_map:
                labeling_fn = labeling_fn_map[labeling_fn]
            else:
                raise ValueError(
                    f"Unrecognized name argument:{labeling_fn}. \
                        Call get_predefined_labeling_functions to \
                            view predefined labeling functions"
                )

        assert callable(labeling_fn), "Labeling function is not callable"

        labeling_function, df, meta = labeling_fn(self._entityset, column_map)

        data = df
        if isinstance(subset, float) or isinstance(subset, int):
            data = data.sample(subset)

        target_entity_index = meta.get("target_entity_index")
        time_index = meta.get("time_index")
        thresh = meta.get("thresh") if thresh is None else thresh
        window_size = meta.get(
            "window_size") if window_size is None else window_size

        label_maker = cp.LabelMaker(
            labeling_function=labeling_function,
            target_dataframe_name=target_entity_index,
            time_index=time_index,
            window_size=window_size,
        )

        kwargs = {**meta, **kwargs}
        kwargs = {
            k: kwargs.get(k)
            for k in set(getfullargspec(label_maker.search)[0])
            if kwargs.get(k) is not None
        }
        label_times = label_maker.search(
            data.sort_values(time_index), num_samples, minimum_data=minimum_data,
            maximum_data=maximum_data, gap=gap, drop_empty=drop_empty, verbose=verbose, **kwargs
        )
        if thresh is not None:
            label_times = label_times.threshold(thresh)

        self._label_times = label_times
        self._label_col_name = "label"
        self._label_times_meta = meta

        return label_times, meta

    @guide
    def set_label_times(self, label_times, label_col_name, meta=None):
        """Set the label times for this Zephyr instance.

        Args:
            label_times (composeml.LabelTimes): Label times.
            label_col_name (str): Name of the label column.
            meta (dict, optional): Additional metadata about the labels.
        """
        assert (isinstance(label_times, cp.LabelTimes))
        self._label_times = label_times
        self._label_col_name = label_col_name
        self._label_times_meta = meta

    @guide
    def get_label_times(self, visualize=False):
        """Get the current label times.

        Args:
            visualize (bool, optional): Whether to display a distribution plot. Defaults to False.

        Returns:
            tuple: (composeml.LabelTimes, dict) The label times and metadata.
        """
        if visualize:
            cp.label_times.plots.LabelPlots(self._label_times).distribution()
        return self._label_times, self._label_times_meta

    @guide
    def generate_feature_matrix(
            self,
            target_dataframe_name=None,
            instance_ids=None,
            agg_primitives=None,
            trans_primitives=None,
            groupby_trans_primitives=None,
            allowed_paths=None,
            max_depth=2,
            ignore_dataframes=None,
            ignore_columns=None,
            primitive_options=None,
            seed_features=None,
            drop_contains=None,
            drop_exact=None,
            where_primitives=None,
            max_features=-1,
            cutoff_time_in_index=False,
            save_progress=None,
            features_only=False,
            training_window=None,
            approximate=None,
            chunk_size=None,
            n_jobs=1,
            dask_kwargs=None,
            verbose=False,
            return_types=None,
            progress_callback=None,
            include_cutoff_time=True,
            add_interesting_values=False,
            max_interesting_values=5,
            interesting_dataframe_name=None,
            interesting_values=None,
            signal_dataframe_name=None,
            signal_column=None,
            signal_transformations=None,
            signal_aggregations=None,
            signal_window_size=None,
            signal_replace_dataframe=False,
            **sigpro_kwargs):
        """Generate a feature matrix using automated feature engineering.
        Note that this method creates a deepcopy
        of the generated or set entityset within the Zephyr instance
        before performing any signal processing or feature generation.

        Args:
            target_dataframe_name (str, optional): Name of target entity for feature engineering.
            instance_ids (list, optional): List of specific instances to generate features for.
            agg_primitives (list, optional): Aggregation primitives to apply.
            trans_primitives (list, optional): Transform primitives to apply.
            groupby_trans_primitives (list, optional): Groupby transform primitives to apply.
            allowed_paths (list, optional): Allowed entity paths for feature generation.
            max_depth (int, optional): Maximum allowed depth of entity relationships.
                Defaults to 2.
            ignore_dataframes (list, optional): Dataframes to ignore during feature generation.
            ignore_columns (dict, optional): Columns to ignore per dataframe.
            primitive_options (dict, optional): Options for specific primitives.
            seed_features (list, optional): Seed features to begin with.
            drop_contains (list, optional): Drop features containing these substrings.
            drop_exact (list, optional): Drop features exactly matching these names.
            where_primitives (list, optional): Primitives to use in where clauses.
            max_features (int, optional): Maximum number of features to return. -1 for all.
            cutoff_time_in_index (bool, optional): Include cutoff time in the index.
            save_progress (str, optional): Path to save progress.
            features_only (bool, optional): Return only features without calculating values.
            training_window (str, optional): Data window to use for training.
            approximate (str, optional): Approximation method to use.
            chunk_size (int, optional): Size of chunks for parallel processing.
            n_jobs (int, optional): Number of parallel jobs. Defaults to 1.
            dask_kwargs (dict, optional): Arguments for dask computation.
            verbose (bool, optional): Whether to display progress. Defaults to False.
            return_types (list, optional): Types of features to return.
            progress_callback (callable, optional): Callback for progress updates.
            include_cutoff_time (bool, optional): Include cutoff time features. Defaults to True.
            add_interesting_values (bool, optional): Add interesting values. Defaults to False.
            max_interesting_values (int, optional): Maximum interesting values per column.
            interesting_dataframe_name (str, optional): Dataframe for interesting values.
            interesting_values (dict, optional): Pre-defined interesting values.
            signal_dataframe_name (str, optional): Name of dataframe containing signal data.
            signal_column (str, optional): Name of column containing signal values.
            signal_transformations (list, optional): Signal transformations to apply.
            signal_aggregations (list, optional): Signal aggregations to apply.
            signal_window_size (str, optional): Window size for signal processing.
            signal_replace_dataframe (bool, optional): Replace original signal dataframe.
            **sigpro_kwargs: Additional arguments for signal processing.

        Returns:
            tuple: (pd.DataFrame, list, featuretools.EntitySet)
                Feature matrix, feature definitions, and the processed entityset.
        """
        entityset_copy = copy.deepcopy(self._entityset)
        # perform signal processing
        if signal_dataframe_name is not None and signal_column is not None:
            # first make copy of entityset
            if signal_transformations is None:
                signal_transformations = []
            if signal_aggregations is None:
                signal_aggregations = []
            process_signals(
                entityset_copy,
                signal_dataframe_name,
                signal_column,
                signal_transformations,
                signal_aggregations,
                signal_window_size,
                signal_replace_dataframe,
                **sigpro_kwargs)

        # add interesting values for where primitives
        if add_interesting_values:
            entityset_copy.add_interesting_values(
                max_values=max_interesting_values,
                verbose=verbose,
                dataframe_name=interesting_dataframe_name,
                values=interesting_values)

        feature_matrix, features = ft.dfs(
            entityset=entityset_copy, cutoff_time=self._label_times,
            target_dataframe_name=target_dataframe_name,
            instance_ids=instance_ids, agg_primitives=agg_primitives,
            trans_primitives=trans_primitives,
            groupby_trans_primitives=groupby_trans_primitives,
            allowed_paths=allowed_paths, max_depth=max_depth,
            ignore_dataframes=ignore_dataframes, ignore_columns=ignore_columns,
            primitive_options=primitive_options, seed_features=seed_features,
            drop_contains=drop_contains, drop_exact=drop_exact,
            where_primitives=where_primitives, max_features=max_features,
            cutoff_time_in_index=cutoff_time_in_index,
            save_progress=save_progress, features_only=features_only,
            training_window=training_window, approximate=approximate,
            chunk_size=chunk_size, n_jobs=n_jobs,
            dask_kwargs=dask_kwargs, verbose=verbose,
            return_types=return_types, progress_callback=progress_callback,
            include_cutoff_time=include_cutoff_time
        )
        self._feature_matrix = self._clean_feature_matrix(
            feature_matrix, label_col_name=self._label_col_name)
        self._features = features

        return self._feature_matrix, self._features, entityset_copy

    @guide
    def get_feature_matrix(self):
        """Get the current feature matrix.

        Returns:
            tuple: (pd.DataFrame, str, list) The feature matrix, label column name,
                and feature definitions.
        """
        return self._feature_matrix, self._label_col_name, self._features

    @guide
    def set_feature_matrix(self, feature_matrix,
                           labels=None, label_col_name="label"):
        """Set the feature matrix for this Zephyr instance.

        Args:
            feature_matrix (pd.DataFrame): The feature matrix to use.
            labels (array-like, optional): Labels to add to the feature matrix.
            label_col_name (str, optional): Name of the label column. Defaults to "label".
        """
        assert isinstance(feature_matrix, pd.DataFrame) and (
            labels is not None or
            label_col_name in feature_matrix.columns
        )
        if labels is not None:
            feature_matrix[label_col_name] = labels
        self._feature_matrix = self._clean_feature_matrix(
            feature_matrix, label_col_name=label_col_name
        )
        self._label_col_name = label_col_name

    @guide
    def generate_train_test_split(
        self,
        test_size=None,
        train_size=None,
        random_state=None,
        shuffle=True,
        stratify=False,
    ):
        """Generate a train-test split of the feature matrix.

        Args:
            test_size (float or int, optional): Proportion or absolute size of test set.
            train_size (float or int, optional): Proportion or absolute size of training set.
            random_state (int, optional): Random seed for reproducibility.
            shuffle (bool, optional): Whether to shuffle before splitting. Defaults to True.
            stratify (bool or list, optional): Whether to maintain label distribution.
                If True, uses labels for stratification. If list, uses those columns.
                Defaults to False.

        Returns:
            tuple: (X_train, X_test, y_train, y_test) The split feature matrices and labels.
        """
        feature_matrix = self._feature_matrix.copy()
        labels = feature_matrix.pop(self._label_col_name)

        if not isinstance(stratify, list):
            if stratify:
                stratify = labels
            else:
                stratify = None

        X_train, X_test, y_train, y_test = train_test_split(
            feature_matrix,
            labels,
            test_size=test_size,
            train_size=train_size,
            random_state=random_state,
            shuffle=shuffle,
            stratify=stratify,
        )

        self._X_train = X_train
        self._X_test = X_test
        self._y_train = y_train
        self._y_test = y_test

        return X_train, X_test, y_train, y_test

    @guide
    def set_train_test_split(self, X_train, X_test, y_train, y_test):
        """Set the train-test split for this Zephyr instance.

        Args:
            X_train (pd.DataFrame): Training features.
            X_test (pd.DataFrame): Testing features.
            y_train (array-like): Training labels.
            y_test (array-like): Testing labels.
        """
        self._X_train = X_train
        self._X_test = X_test
        self._y_train = y_train
        self._y_test = y_test

    @guide
    def get_train_test_split(self):
        """Get the current train-test split.

        Returns:
            tuple or None: (X_train, X_test, y_train, y_test) if split exists, None otherwise.
        """
        if (self._X_train is None or self._X_test is None or
                self._y_train is None or self._y_test is None):
            return None
        return self._X_train, self._X_test, self._y_train, self._y_test

    @guide
    def set_fitted_pipeline(self, pipeline):
        """Set a fitted pipeline for this Zephyr instance.

        Args:
            pipeline (MLPipeline): The fitted pipeline to use.
        """
        self._pipeline = pipeline

    @guide
    def fit_pipeline(
            self, pipeline="xgb_classifier", pipeline_hyperparameters=None,
            X=None, y=None, visual=False, **kwargs):
        """Fit a machine learning pipeline.

        Args:
            pipeline (str or dict or MLPipeline, optional): Pipeline to use. Can be:
                - Name of a registered pipeline (default: "xgb_classifier")
                - Path to a JSON pipeline specification
                - Dictionary with pipeline specification
                - MLPipeline instance
            pipeline_hyperparameters (dict, optional): Hyperparameters for the pipeline.
            X (pd.DataFrame, optional): Training features. If None, uses stored training set.
            y (array-like, optional): Training labels. If None, uses stored training labels.
            visual (bool, optional): Whether to return visualization data. Defaults to False.
            **kwargs: Additional arguments passed to the pipeline's fit method.

        Returns:
            dict or None: If visual=True, returns visualization data dictionary.
        """
        self._pipeline = self._get_mlpipeline(
            pipeline, pipeline_hyperparameters)

        if X is None:
            X = self._X_train
        if y is None:
            y = self._y_train

        if visual:
            outputs_spec, visual_names = self._get_outputs_spec(False)
        else:
            outputs_spec = None

        outputs = self._pipeline.fit(X, y, output_=outputs_spec, **kwargs)

        if visual and outputs is not None:
            return dict(zip(visual_names, outputs))

    @guide
    def get_fitted_pipeline(self):
        """Get the current fitted pipeline.

        Returns:
            MLPipeline: The current fitted pipeline.
        """
        return self._pipeline

    @guide
    def predict(self, X=None, visual=False, **kwargs):
        """Make predictions using the fitted pipeline.

        Args:
            X (pd.DataFrame, optional): Features to predict on. If None, uses test set.
            visual (bool, optional): Whether to return visualization data. Defaults to False.
            **kwargs: Additional arguments passed to the pipeline's predict method.

        Returns:
            array-like or tuple: Predictions, and if visual=True, also returns visualization data.
        """
        if X is None:
            X = self._X_test
        if visual:
            outputs_spec, visual_names = self._get_outputs_spec()
        else:
            outputs_spec = "default"

        outputs = self._pipeline.predict(X, output_=outputs_spec, **kwargs)
        if visual and visual_names:
            prediction = outputs[0]
            return prediction, dict(
                zip(visual_names, outputs[-len(visual_names):]))

        return outputs

    @guide
    def evaluate(
            self, X=None, y=None, metrics=None, global_args=None,
            local_args=None, global_mapping=None, local_mapping=None):
        """Evaluate the fitted pipeline's performance.

        Args:
            X (pd.DataFrame, optional): Features to evaluate on. If None, uses test set.
            y (array-like, optional): True labels. If None, uses test labels.
            metrics (list, optional): Metrics to compute. If None, uses DEFAULT_METRICS.
            global_args (dict, optional): Arguments passed to all metrics.
            local_args (dict, optional): Arguments passed to specific metrics.
            global_mapping (dict, optional): Mapping applied to all metric inputs.
            local_mapping (dict, optional): Mapping applied to specific metric inputs.

        Returns:
            dict: A dictionary mapping metric names to their computed values.
        """
        if X is None:
            X = self._X_test
        if y is None:
            y = self._y_test

        final_context = self._pipeline.predict(X, output_=-1)

        # remap items, if any
        if global_mapping is not None:
            for cur, new in global_mapping.items():
                if cur in final_context:
                    cur_item = final_context.pop(cur)
                    final_context[new] = cur_item

        if metrics is None:
            metrics = DEFAULT_METRICS

        if global_args is None:
            global_args = {}

        if local_args is None:
            local_args = {}

        if local_mapping is None:
            local_mapping = {}

        results = {}
        for metric in metrics:
            try:
                metric_primitive = self._get_ml_primitive(metric)

                if metric in local_mapping:
                    metric_context = {}
                    metric_mapping = local_mapping[metric]
                    for cur, item in final_context.items():
                        new = metric_mapping.get(cur, cur)
                        metric_context[new] = item
                else:
                    metric_context = final_context

                if metric in local_args:
                    metric_args = local_args[metric]
                else:
                    metric_args = {}

                res = metric_primitive.produce(
                    y_true=self._y_test, **metric_context, **metric_args)
                results[metric_primitive.name] = res
            except Exception as e:
                LOGGER.error(
                    f"Unable to run evaluation metric: {metric_primitive.name}",
                    exc_info=e)
        self._results = results
        return results

    def _clean_feature_matrix(self, feature_matrix, label_col_name="label"):
        labels = feature_matrix.pop(label_col_name)

        count_cols = feature_matrix.filter(like="COUNT").columns
        feature_matrix[count_cols] = feature_matrix[count_cols].apply(
            lambda x: x.astype(np.int64)
        )

        string_cols = feature_matrix.select_dtypes(include="category").columns
        feature_matrix = pd.get_dummies(feature_matrix, columns=string_cols)

        feature_matrix[label_col_name] = labels

        return feature_matrix

    def _get_mlpipeline(self, pipeline, hyperparameters=None):
        if isinstance(pipeline, str) and os.path.isfile(pipeline):
            with open(pipeline) as json_file:
                pipeline = json.load(json_file)

        mlpipeline = MLPipeline(pipeline)
        if hyperparameters:
            mlpipeline.set_hyperparameters(hyperparameters)

        return mlpipeline

    def _get_ml_primitive(self, primitive, hyperparameters=None):
        if isinstance(primitive, str) and os.path.isfile(primitive):
            with open(primitive) as json_file:
                primitive = json.load(json_file)
        mlprimitive = MLBlock(primitive)

        if hyperparameters:
            mlprimitive.set_hyperparameters(hyperparameters)
        return mlprimitive

    def _get_outputs_spec(self, default=True):
        outputs_spec = ["default"] if default else []

        try:
            visual_names = self._pipeline.get_output_names("visual")
            outputs_spec.append("visual")
        except ValueError:
            visual_names = []

        return outputs_spec, visual_names


if __name__ == "__main__":
    obj = Zephyr()
    print(obj.GET_EVALUATION_METRICS())
    alarms_df = pd.DataFrame(
        {
            "COD_ELEMENT": [0, 0],
            "DAT_START": [
                pd.Timestamp("2022-01-01 00:00:00"),
                pd.Timestamp("2022-03-01 11:12:13"),
            ],
            "DAT_END": [
                pd.Timestamp("2022-01-01 13:00:00"),
                pd.Timestamp("2022-03-02 11:12:13"),
            ],
            "IND_DURATION": [0.5417, 1.0],
            "COD_ALARM": [12345, 98754],
            "COD_ALARM_INT": [12345, 98754],
            "DES_NAME": ["Alarm1", "Alarm2"],
            "DES_TITLE": ["Description of alarm 1", "Description of alarm 2"],
        }
    )
    stoppages_df = pd.DataFrame(
        {
            "COD_ELEMENT": [0, 0],
            "DAT_START": [
                pd.Timestamp("2022-01-01 00:00:00"),
                pd.Timestamp("2022-03-01 11:12:13"),
            ],
            "DAT_END": [
                pd.Timestamp("2022-01-08 11:07:17"),
                pd.Timestamp("2022-03-01 17:00:13"),
            ],
            "DES_WO_NAME": ["stoppage name 1", "stoppage name 2"],
            "DES_COMMENTS": ["description of stoppage 1", "description of stoppage 2"],
            "COD_WO": [12345, 67890],
            "IND_DURATION": [7.4642, 0.2417],
            "IND_LOST_GEN": [45678.0, 123.0],
            "COD_ALARM": [12345, 12345],
            "COD_CAUSE": [32, 48],
            "COD_INCIDENCE": [987654, 123450],
            "COD_ORIGIN": [6, 23],
            "COD_STATUS": ["STOP", "PAUSE"],
            "COD_CODE": ["ABC", "XYZ"],
            "DES_DESCRIPTION": ["Description 1", "Description 2"],
        }
    )
    notifications_df = pd.DataFrame(
        {
            "COD_ELEMENT": [0, 0],
            "COD_ORDER": [12345, 67890],
            "IND_QUANTITY": [1, -20],
            "COD_MATERIAL_SAP": [36052411, 67890],
            "DAT_POSTING": [
                pd.Timestamp("2022-01-01 00:00:00"),
                pd.Timestamp("2022-03-01 00:00:00"),
            ],
            "COD_MAT_DOC": [77889900, 12345690],
            "DES_MEDIUM": [
                "Description of notification 1",
                "Description of notification 2",
            ],
            "COD_NOTIF": [567890123, 32109877],
            "DAT_MALF_START": [
                pd.Timestamp("2021-12-25 18:07:10"),
                pd.Timestamp("2022-02-28 06:04:00"),
            ],
            "DAT_MALF_END": [
                pd.Timestamp("2022-01-08 11:07:17"),
                pd.Timestamp("2022-03-01 17:00:13"),
            ],
            "IND_BREAKDOWN_DUR": [14.1378, 2.4792],
            "FUNCT_LOC_DES": ["location description 1", "location description 2"],
            "COD_ALARM": [12345, 12345],
            "DES_ALARM": ["Alarm description", "Alarm description"],
        }
    )
    work_orders_df = pd.DataFrame(
        {
            "COD_ELEMENT": [0, 0],
            "COD_ORDER": [12345, 67890],
            "DAT_BASIC_START": [
                pd.Timestamp("2022-01-01 00:00:00"),
                pd.Timestamp("2022-03-01 00:00:00"),
            ],
            "DAT_BASIC_END": [
                pd.Timestamp("2022-01-09 00:00:00"),
                pd.Timestamp("2022-03-02 00:00:00"),
            ],
            "COD_EQUIPMENT": [98765, 98765],
            "COD_MAINT_PLANT": ["ABC", "ABC"],
            "COD_MAINT_ACT_TYPE": ["XYZ", "XYZ"],
            "COD_CREATED_BY": ["A1234", "B6789"],
            "COD_ORDER_TYPE": ["A", "B"],
            "DAT_REFERENCE": [
                pd.Timestamp("2022-01-01 00:00:00"),
                pd.Timestamp("2022-03-01 00:00:00"),
            ],
            "DAT_CREATED_ON": [
                pd.Timestamp("2022-03-01 00:00:00"),
                pd.Timestamp("2022-04-18 00:00:00"),
            ],
            "DAT_VALID_END": [pd.NaT, pd.NaT],
            "DAT_VALID_START": [pd.NaT, pd.NaT],
            "COD_SYSTEM_STAT": ["ABC XYZ", "LMN OPQ"],
            "DES_LONG": ["description of work order", "description of work order"],
            "COD_FUNCT_LOC": ["!12345", "?09876"],
            "COD_NOTIF_OBJ": ["00112233", "00998877"],
            "COD_MAINT_ITEM": ["", "019283"],
            "DES_MEDIUM": ["short description", "short description"],
            "DES_FUNCT_LOC": ["XYZ1234", "ABC9876"],
        }
    )
    turbines_df = pd.DataFrame(
        {
            "COD_ELEMENT": [0],
            "TURBINE_PI_ID": ["TA00"],
            "TURBINE_LOCAL_ID": ["A0"],
            "TURBINE_SAP_COD": ["LOC000"],
            "DES_CORE_ELEMENT": ["T00"],
            "SITE": ["LOCATION"],
            "DES_CORE_PLANT": ["LOC"],
            "COD_PLANT_SAP": ["ABC"],
            "PI_COLLECTOR_SITE_NAME": ["LOC0"],
            "PI_LOCAL_SITE_NAME": ["LOC0"],
        }
    )
    pidata_df = pd.DataFrame(
        {
            "time": [
                pd.Timestamp("2022-01-02 13:21:01"),
                pd.Timestamp("2022-03-08 13:21:01"),
            ],
            "COD_ELEMENT": [0, 0],
            "val1": [9872.0, 559.0],
            "val2": [10.0, -7.0],
        }
    )

    # obj.create_entityset(
    #     {
    #         "alarms": alarms_df,
    #         "stoppages": stoppages_df,
    #         "notifications": notifications_df,
    #         "work_orders": work_orders_df,
    #         "turbines": turbines_df,
    #         "pidata": pidata_df,
    #     },
    #     "pidata",
    # )

    # obj.set_entityset(entityset_path =
    # "/Users/raymondpan/zephyr/Zephyr-repo/brake_pad_es", es_type = 'scada')

    # obj.set_labeling_function(name="brake_pad_presence")

    # obj.generate_label_times(labeling_fn="brake_pad_presence",
    # num_samples=10, gap="20d")
    # # print(obj.get_label_times())

    # obj.generate_feature_matrix_and_labels(
    #     target_dataframe_name="turbines",
    #     cutoff_time_in_index=True,
    #     agg_primitives=["count", "sum", "max"],
    #     verbose = True
    # )

    # print(obj.get_feature_matrix_and_labels)

    # obj.generate_train_test_split()
    # add_primitives_path(
    #     path="/Users/raymondpan/zephyr/Zephyr-repo/zephyr_ml/primitives/jsons"
    # )
    # obj.set_and_fit_pipeline()

    # obj.evaluate()




from zephyr_ml.labeling.utils import denormalize


def total_power_loss(es, column_map={}):
    """Calculates the total power loss over the data slice.

    Args:
        es (ft.EntitySet):
            EntitySet of data to calculate total power loss across.
        column_map (dict):
            Optional dictionary to update default column names to the
            actual corresponding column names in the data slice. Can contain the
            following keys:
                "lost_gen": Column that contains the generation lost due to stoppage. Defaults
                    to "IND_LOST_GEN".
                "turbine_id": Column containing the ID of the turbine associated with a
                    stoppage. Must match the index column of the 'turbines' entity.
                    Defaults to "COD_ELEMENT".
                "time_index": Column to use as the time index for the data slice.
                    Defaults to "DAT_START".

    Returns:
        label:
            Labeling function to find the total power loss over a data slice.
        df:
            Denormalized dataframe of data to get labels from
        meta:
            Dictionary containing metadata about labeling function

    """
    lost_gen = column_map.get('lost_gen', 'IND_LOST_GEN')
    turbine_id = column_map.get('turbine_id_column', 'COD_ELEMENT')
    time_index = column_map.get('time_index', 'DAT_START')

    def label(ds, **kwargs):
        return sum(ds[lost_gen])

    meta = {
        "target_entity_index": turbine_id,
        "time_index": time_index,
    }

    df = denormalize(es, entities=['stoppages'])

    return label, df, meta



import copy

# Default EntitySet keyword arguments for entities
DEFAULT_ES_KWARGS = {
    "alarms": {
        "index": "_index",
        "make_index": True,
        "time_index": "DAT_START",
        "secondary_time_index": {"DAT_END": ["IND_DURATION"]},
        "logical_types": {
            "COD_ELEMENT": "categorical",  # turbine id
            "DAT_START": "datetime",  # start
            "DAT_END": "datetime",  # end
            "IND_DURATION": "double",  # duration
            "COD_ALARM": "categorical",  # alarm code
            "COD_ALARM_INT": "categorical",  # international alarm code
            "DES_NAME": "categorical",  # alarm name
            "DES_TITLE": "categorical",  # alarm description
            "COD_STATUS": "categorical",  # status code
        },
    },
    "stoppages": {
        "index": "_index",
        "make_index": True,
        "time_index": "DAT_START",
        "secondary_time_index": {"DAT_END": ["IND_DURATION", "IND_LOST_GEN"]},
        "logical_types": {
            "COD_ELEMENT": "categorical",  # turbine id
            "DAT_START": "datetime",  # start
            "DAT_END": "datetime",  # end
            "DES_WO_NAME": "natural_language",  # work order name
            "DES_COMMENTS": "natural_language",  # work order comments
            "COD_WO": "integer_nullable",  # stoppage code
            "IND_DURATION": "double",  # duration
            "IND_LOST_GEN": "double",  # generation loss
            "COD_ALARM": "categorical",  # alarm code
            "COD_CAUSE": "categorical",  # stoppage cause
            "COD_INCIDENCE": "categorical",  # incidence code
            "COD_ORIGIN": "categorical",  # origin code
            "DESC_CLASS": "categorical",  # ????
            "COD_STATUS": "categorical",  # status code
            "COD_CODE": "categorical",  # stoppage code
            "DES_DESCRIPTION": "natural_language",  # stoppage description
            "DES_TECH_NAME": "categorical",  # turbine technology
        },
    },
    "notifications": {
        "index": "_index",
        "make_index": True,
        "time_index": "DAT_POSTING",
        "secondary_time_index": {"DAT_MALF_END": ["IND_BREAKDOWN_DUR"]},
        "logical_types": {
            "COD_ELEMENT": "categorical",  # turbine id
            "COD_ORDER": "categorical",
            "IND_QUANTITY": "double",
            "COD_MATERIAL_SAP": "categorical",
            "DAT_POSTING": "datetime",
            "COD_MAT_DOC": "categorical",
            "DES_MEDIUM": "categorical",
            "COD_NOTIF": "categorical",
            "DAT_MALF_START": "datetime",
            "DAT_MALF_END": "datetime",
            "IND_BREAKDOWN_DUR": "double",
            "FUNCT_LOC_DES": "categorical",
            "COD_ALARM": "categorical",
            "DES_ALARM": "categorical",
        },
    },
    "work_orders": {
        "index": "COD_ORDER",
        "time_index": "DAT_BASIC_START",
        "secondary_time_index": {"DAT_VALID_END": []},
        "logical_types": {
            "COD_ELEMENT": "categorical",
            "COD_ORDER": "categorical",
            "DAT_BASIC_START": "datetime",
            "DAT_BASIC_END": "datetime",
            "COD_EQUIPMENT": "categorical",
            "COD_MAINT_PLANT": "categorical",
            "COD_MAINT_ACT_TYPE": "categorical",
            "COD_CREATED_BY": "categorical",
            "COD_ORDER_TYPE": "categorical",
            "DAT_REFERENCE": "datetime",
            "DAT_CREATED_ON": "datetime",
            "DAT_VALID_END": "datetime",
            "DAT_VALID_START": "datetime",
            "COD_SYSTEM_STAT": "categorical",
            "DES_LONG": "natural_language",
            "COD_FUNCT_LOC": "categorical",
            "COD_NOTIF_OBJ": "categorical",
            "COD_MAINT_ITEM": "categorical",
            "DES_MEDIUM": "natural_language",
            "DES_FUNCT_LOC": "categorical",
        },
    },
    "turbines": {
        "index": "COD_ELEMENT",
        "logical_types": {
            "COD_ELEMENT": "categorical",
            "TURBINE_PI_ID": "categorical",
            "TURBINE_LOCAL_ID": "categorical",
            "TURBINE_SAP_COD": "categorical",
            "DES_CORE_ELEMENT": "categorical",
            "SITE": "categorical",
            "DES_CORE_PLANT": "categorical",
            "COD_PLANT_SAP": "categorical",
            "PI_COLLECTOR_SITE_NAME": "categorical",
            "PI_LOCAL_SITE_NAME": "categorical",
        },
    },
}

DEFAULT_ES_TYPE_KWARGS = {
    "pidata": {
        "index": "_index",
        "make_index": True,
        "time_index": "time",
        "logical_types": {"time": "datetime", "COD_ELEMENT": "categorical"},
    },
    "scada": {
        "index": "_index",
        "make_index": True,
        "time_index": "TIMESTAMP",
        "logical_types": {"TIMESTAMP": "datetime", "COD_ELEMENT": "categorical"},
    },
    "vibrations": {
        "index": "_index",
        "make_index": True,
        "time_index": "timestamp",
        "logical_types": {
            "COD_ELEMENT": "categorical",
            "turbine_id": "categorical",
            "signal_id": "categorical",
            "timestamp": "datetime",
            "sensorName": "categorical",
            "sensorType": "categorical",
            "sensorSerial": "integer_nullable",
            "siteName": "categorical",
            "turbineName": "categorical",
            "turbineSerial": "integer_nullable",
            "configurationName": "natural_language",
            "softwareVersion": "categorical",
            "rpm": "double",
            "rpmStatus": "natural_language",
            "duration": "natural_language",
            "condition": "categorical",
            "maskTime": "datetime",
            "Mask Status": "natural_language",
            "System Serial": "categorical",
            "WPS-ActivePower-Average": "double",
            "WPS-ActivePower-Minimum": "double",
            "WPS-ActivePower-Maximum": "double",
            "WPS-ActivePower-Deviation": "double",
            "WPS-ActivePower-StartTime": "datetime",
            "WPS-ActivePower-StopTime": "datetime",
            "WPS-ActivePower-Counts": "natural_language",
            "Measured RPM": "double",
            "WPS-ActivePower": "double",
            "WPS-Gearoiltemperature": "double",
            "WPS-GeneratorRPM": "double",
            "WPS-PitchReference": "double",
            "WPS-RotorRPM": "double",
            "WPS-Windspeed": "double",
            "WPS-YawAngle": "double",
            "overload warning": "categorical",
            "bias warning": "categorical",
            "bias voltage": "double",
            "xValueOffset": "double",
            "xValueDelta": "double",
            "xValueUnit": "categorical",
            "yValueUnit": "categorical",
            "TotalCount-RPM0": "double",
            "TotalCount-RPM1": "double",
            "TotalCount-RPM2": "double",
            "TotalCount-RPM3": "double",
        },
    },
}


def get_mapped_kwargs(es_type, new_kwargs=None):
    if es_type not in DEFAULT_ES_TYPE_KWARGS.keys():
        raise ValueError("Unrecognized es_type argument: {}".format(es_type))
    mapped_kwargs = DEFAULT_ES_KWARGS.copy()
    mapped_kwargs.update({es_type: DEFAULT_ES_TYPE_KWARGS[es_type]})

    if new_kwargs is not None:
        if not isinstance(new_kwargs, dict):
            raise ValueError(
                "new_kwargs must be dictionary mapping entity name to dictionary "
                "with updated keyword arguments for EntitySet creation."
            )
        for entity in new_kwargs:
            if entity not in mapped_kwargs:
                raise ValueError(
                    'Unrecognized entity "{}" found in new keyword argument '
                    "mapping.".format(entity)
                )

            mapped_kwargs[entity].update(new_kwargs[entity])
    return mapped_kwargs


def get_default_es_type_kwargs():
    return copy.deepcopy(DEFAULT_ES_TYPE_KWARGS)


def get_es_types():
    return DEFAULT_ES_TYPE_KWARGS.keys()




import numpy as np
import pandas as pd


def _search_relationship(es, left, right):
    for r in es.relationships:
        if r.parent_name in left:
            if right == r.child_name:
                left_on = r.parent_column.name
                right_on = r.child_column.name

        elif r.child_name in left:
            if right == r.parent_name:
                left_on = r.child_column.name
                right_on = r.parent_column.name

    return left_on, right_on


def denormalize(es, entities):
    """Merge a set of entities into a single dataframe.
    Convert a set of entities from the entityset into a single
    dataframe by repetitively merging the selected entities. The
    merge process is applied sequentially.
    Args:
        entities (list):
            list of strings denoting which entities to merge.
    Returns:
        pandas.DataFrame:
            A single dataframe containing all the information from the
            selected entities.
    """
    k = len(entities)

    # initial entity to start from (should be the target entity)
    first = entities[0]
    previous = [first]
    df = es[first]

    # merge the dataframes to create a single input
    for i in range(1, k):
        right = entities[i]

        left_on, right_on = _search_relationship(es, previous, right)
        df = pd.merge(df, es[right],
                      left_on=left_on, right_on=right_on,
                      how='left', suffixes=('', '_y')).filter(regex='^(?!.*_y)')

        previous.append(right)

    return df


def required_columns(columns):
    """Decorator function for recording required columns for a function."""
    def wrapper(wrapped):
        def func(*args, **kwargs):
            return wrapped(*args, **kwargs)

        func.__required_columns__ = columns
        func.__doc__ = wrapped.__doc__
        func.__name__ = wrapped.__name__
        return func

    return wrapper


def merge_binary_labeling_functions(labeling_functions, and_connected=True):
    """Generates a labeling function from merging multiple binary labeling functions.

    Args:
        labeling_functions (list):
            A list of labeling functions (with df as an input) to merge.
        and_connected (bool):
            If and_connected is True, each individual labeling function criteria must be True
            for the output function to give a positive label. If and_connected is False,
            at least one labeling function criteria has to be met for the output function
            to give a positive label. Default is True.

    Returns:
        function:
            A function that takes in a dataframe, which is derived from the input labeling
            functions.
    """
    def merged_function(df):
        out = and_connected
        for function in labeling_functions:
            if and_connected:
                out &= function(df)
            else:
                out |= function(df)

        return int(out)

    return merged_function


def aggregate_by_column(numerical_column, aggregation):
    """Generates a function for aggregates numerical column values over a data slice.

    Args:
        numerical_column (str):
            Numerical column to aggregate over.
        aggregation (function):
            Aggregation function to apply.

    Returns:
        function:
            The function returns the total numerical column value over the data
            slice as a continuous label.
    """
    def aggregate_function(df):
        """Aggregate function with:
        numerical_column={}
        aggregation={}
        """
        return aggregation(df[numerical_column])

    aggregate_function.__doc__ = aggregate_function.__doc__.format(numerical_column,
                                                                   aggregation.__name__)

    return aggregate_function


def categorical_presence(categorical_column, value):
    """Generates a function that determines if the categorical column has the desired value.

    Args:
        categorical_column (str):
            Categorical column to use values from.
        value (str or int or float):
            Value to compare categorical columns values to.

    Returns:
        function:
            The function returns 1 if categorical column has the desired value,
            0 otherwise.
    """
    def categorical_function(df):
        """Categorical presence function with:
        categorical_column={}
        value={}
        """
        return int(df[categorical_column].isin([value]).sum() > 0)

    categorical_function.__doc__ = categorical_function.__doc__.format(
        categorical_column, value)
    return categorical_function


def keyword_in_text(keyword, columns=None):
    """Determines presence of keyword in text field data columns.

    Args:
        keyword (str):
            Keyword to search the text columns for.
        columns (list or None):
            List of columns to search through to find keyword. If None, all
            columns are tested. Default is None.

    Returns:
        function:
            The function returns 1 if the keyword is present in any column,
            0 otherwise.
    """
    def keyword_function(df):
        """Keyword function with:
        keyword={}
        columns={}
        """
        mask = np.full(len(df), False)
        for col in columns:
            try:
                mask |= df[col].str.contains(keyword, case=False, na=False)
            except KeyError:
                print("Unable to find column for keyword search")

        return int(mask.sum() != 0)

    keyword_function.__doc__ = keyword_function.__doc__.format(
        keyword, columns)
    return keyword_function


def greater_than(numerical_column, threshold):
    """Generates a function to see if there are numerical values greater than a threshold.

    Args:
        numerical_column (str):
            Numerical column to use values from.
        threshold (float):
            Threshold for the numerical values used to define the binary labels.

    Returns:
        function:
            The function returns 1 if data contains a value is greater than threshold,
            0 otherwise.
    """
    def numerical_function(df):
        """Numerical presence function with:
        numerical_column={}
        threshold={}
        """
        series = df[numerical_column]
        return int(len(series[series > threshold]) > 0)

    numerical_function.__doc__ = numerical_function.__doc__.format(
        numerical_column, threshold)
    return numerical_function


def total_duration(start_time, end_time):
    """Generates function for calculating the total duration given start/end time indexes.

    Args:
        start_time (str):
            Name of the start time column.
        end_time (str):
            Name of the end time column.

    Returns:
        function:
            The function returns the total duration in seconds based on the two
            given time endpoints for the data slice.
    """
    def duration_function(df):
        """Duration function with:
        start_time={}
        end_time={}
        """
        return ((df[end_time] - df[start_time]).dt.total_seconds()).sum()

    duration_function.__doc__ = duration_function.__doc__.format(
        start_time, end_time)
    return duration_function



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for zephyr_ml package."""
from datetime import datetime, time

import numpy as np
import pandas as pd

STOPPAGES_DATA = {
    'stoppage_id': [0, 1],
    'WTG': ['A001', 'A001'],
    'Start': [pd.Timestamp('2019-08-01 00:00:00.123000'),
              pd.Timestamp('2019-09-01 00:00:00.456000')],
    'End': [pd.Timestamp('2019-08-23 23:59:59.123000'),
            pd.Timestamp('2019-09-19 01:02:03.456000')],
    'Duration (hh:mi:ss)': [datetime(1900, 1, 30, 3, 4, 5),
                            time(14, 15, 16)],
    'Status': [np.nan, np.nan],
    'Cause': ['Cause', 'Cause'],
    'Origin': [np.nan, np.nan],
    'Class': [np.nan, np.nan],
    'Alarm': [np.nan, np.nan],
    'Est. Lost Gen (kWh)': [1234.0, 45678.0],
    'Comments': ['Yaw Error Stoppage', 'Comment'],
    'SAP W. Order': [np.nan, np.nan],
    'SAP W. Order Name': [np.nan, np.nan],
    'Alarms CORE': [' ', ' ']
}

WORK_ORDERS_DATA = {
    'Order reference timestamp': [pd.Timestamp('2019-07-27 23:59:59'),
                                  pd.Timestamp('2019-08-02 13:21:01')],
    'Order basic start timestamp': [pd.Timestamp('2019-07-27 12:09:41'),
                                    pd.Timestamp('2019-08-02 03:04:56')],
    'Order basic finish timestamp': [pd.Timestamp('2019-08-04 09:08:13'),
                                     pd.Timestamp('2019-08-28 01:28:23')],
    'Order': [123456789.0, 987654321.0],
    'Order created on date': [pd.Timestamp('2019-07-27 00:00:00'),
                              pd.Timestamp('2019-08-02 00:00:00')],
    'Order actual release date': [pd.Timestamp('2019-07-27 00:00:00'),
                                  pd.Timestamp('2019-08-02 00:00:00')],
    'Order actual start date': [pd.Timestamp(np.nan), pd.Timestamp('2019-04-21 00:00:00')],
    'Order actual finish date': [pd.Timestamp(np.nan), pd.Timestamp(np.nan)],
    'Maintenance activity type': ['MPI', 'MPI'],
    'Maintenance activity description': ['Yaw Maintenance', 'Description'],
    'Functional location': ['1234A001123', '2345A001ABC'],
    'Functional location description': ['Func Description 1', 'Func Description 2'],
    'Notification': [1000101115.0, 1000201113.0]
}

NOTIFICATIONS_DATA = {
    'Notification timestamp': [pd.Timestamp('2019-07-28 02:12:34'),
                               pd.Timestamp('2019-08-02 14:12:34')],
    'Malfunction start timestamp': [pd.Timestamp('2019-07-29 00:01:23'),
                                    pd.Timestamp('2019-08-03 00:04:56')],
    'Malfunction end timestamp': [pd.Timestamp('2019-08-03 11:44:23'),
                                  pd.Timestamp('2019-08-20 10:22:16')],
    'Notification created timestamp': [pd.Timestamp('2019-07-30 03:12:34'),
                                       pd.Timestamp('2019-08-04 15:12:34')],
    'Notification changed timestamp': [pd.Timestamp('2019-09-03 05:06:07'),
                                       pd.Timestamp('2019-10-20 05:06:07')],
    'Required start timestamp': [pd.Timestamp('2019-07-28 02:27:06'),
                                 pd.Timestamp('2019-08-02 14:30:44')],
    'Required end timestamp': [pd.Timestamp('2019-07-30 10:19:12'),
                               pd.Timestamp('2019-08-04 11:45:14')],
    'Completion timestamp': [pd.Timestamp('2019-09-03 04:05:06'),
                             pd.Timestamp('2019-10-20 04:05:06')],
    'Order': [123456789.0, 987654321.0],
    'Notification': [1000101115.0, 1000201113.0],
    'Breakdown (X=yes)': ['X', 'nan'],
    'Breakdown duration (hrs)': [12.0, 13.0],
    'Technical object description': ['nan', 'nan'],
    'Material description': ['nan', 'nan'],
    'Assembly description': ['nan', 'nan'],
    'Long text available (X=yes)': ['nan', 'nan'],
    'Subject long text': ['nan', 'nan'],
    'Equipment number': [np.nan, np.nan],
    'Assembly number': [np.nan, np.nan],
    'Priority code': ['P', 'P'],
    'Priority': ['Priority', 'Priority'],
    'Serial number': ['nan', 'nan'],
    'Material number': [np.nan, np.nan],
    'Causing element description': ['nan', 'nan'],
    'Fault mode description': ['nan', 'nan'],
    'Cause of malfunction description': ['nan', 'nan'],
    'Subject description': ['nan', 'Description'],
    'Functional location': ['1234A001123', '2345A001ABC'],
    'Functional location description': ['Func Description 1', 'Func Description 2'],
}


def merge_work_orders_notifications_data():
    """Helper function to merge work orders and notifications data."""
    changed_wo_data = WORK_ORDERS_DATA.copy()
    changed_wo_data['WTG'] = ['A001', 'A001']
    changed_notif_data = NOTIFICATIONS_DATA.copy()
    # matching the output of the merge
    changed_notif_data['Functional location_y'] = changed_notif_data.pop(
        'Functional location')
    changed_notif_data['Functional location description_y'] = (
        changed_notif_data.pop('Functional location description'))
    # matching the notifications update
    changed_wo_data.update(changed_notif_data)
    return changed_wo_data


def merge_label_generation_data():
    expected_data = STOPPAGES_DATA.copy()
    expected_data['stoppage_id'] = [0, 1]
    expected_won = merge_work_orders_notifications_data()

    for key, value in expected_won.items():
        if key not in expected_data:
            expected_data[key] = [expected_won[key][1], np.nan]

    return expected_data



from zephyr_ml.core import GuideHandler, guide


class DummyObject:
    def __init__(self):
        producers_and_getters = [
            ([self.step0_key, self.step0_set], [self.step0_getter]),
            ([self.step1_key, self.step1_set], [self.step1_getter]),
            ([self.step2_key, self.step2_set], [self.step2_getter])
        ]
        set_methods = {
            self.step0_set.__name__,
            self.step1_set.__name__,
            self.step2_set.__name__
        }
        self._guide_handler = GuideHandler(producers_and_getters, set_methods)

    @guide
    def step0_key(self):
        return "step0_key_result"

    @guide
    def step0_set(self):
        return "step0_set_result"

    @guide
    def step0_getter(self):
        return "step0_get_result"

    @guide
    def step1_key(self):
        return "step1_key_result"

    @guide
    def step1_set(self):
        return "step1_set_result"

    @guide
    def step1_getter(self):
        return "step1_get_result"

    @guide
    def step2_key(self):
        return "step2_key_result"

    @guide
    def step2_set(self):
        return "step2_set_result"

    @guide
    def step2_getter(self):
        return "step2_get_result"


def test_forward_key_steps():
    """Test performing key steps in forward order"""
    obj = DummyObject()

    # First step should work without warnings
    assert obj.step0_key() == "step0_key_result"

    # Second step should work without warnings since previous step is up to
    # date
    assert obj.step1_key() == "step1_key_result"

    # Third step should work without warnings since previous step is up to date
    assert obj.step2_key() == "step2_key_result"


def test_set_methods_can_skip(caplog):
    """Test that set methods can skip steps"""
    obj = DummyObject()

    # Set methods should work in any order and start new iterations
    assert obj.step2_set() == "step2_set_result"  # Skip to step 2
    assert "[GUIDE] STALE WARNING" in caplog.text

    assert obj.step0_set() == "step0_set_result"  # Go back to step 0
    assert "[GUIDE] STALE WARNING" in caplog.text
    assert obj.step1_set() == "step1_set_result"  # Do step 1
    assert "[GUIDE] STALE WARNING" in caplog.text


def test_key_methods_require_previous_step(caplog):
    """Test that key methods require the previous step to be up to date"""
    obj = DummyObject()

    # Try to do step 1 without doing step 0 first
    obj.step1_key()
    assert "[GUIDE] INCONSISTENCY WARNING" in caplog.text

    # Do step 0, then step 1 should work
    obj.step0_key()
    assert obj.step1_key() == "step1_key_result"


def test_stale_data_warning(caplog):
    """Test warning when data becomes stale"""
    obj = DummyObject()

    # Complete steps 0 and 1
    obj.step0_key()
    obj.step1_key()

    # Go back to step 0 with set method (allowed, but warns about stale data)
    obj.step0_set()
    assert "[GUIDE] STALE WARNING" in caplog.text


def test_getter_with_stale_data(caplog):
    """Test getting data that may be stale"""
    obj = DummyObject()

    # Complete steps 0 and 1
    obj.step0_key()
    obj.step1_key()

    # Go back to step 0 with set method
    obj.step0_set()

    # Try to get data from step 1, should warn about stale data
    obj.step1_getter()
    assert "[GUIDE] STALE WARNING" in caplog.text


def test_getter_with_missing_key(caplog):
    """Test getting data when the key method hasn't been run"""
    obj = DummyObject()

    # Try to get data without running key method first
    obj.step1_getter()
    assert "[GUIDE] INCONSISTENCY WARNING" in caplog.text


def test_key_method_after_stale_data(caplog):
    """Test that key methods cannot be run when previous step is stale"""
    obj = DummyObject()

    # Complete steps 0 and 1
    obj.step0_key()
    obj.step1_key()

    # Go back to step 0 with set method
    obj.step1_set()
    obj.step1_key()
    assert "[GUIDE] INCONSISTENCY WARNING" in caplog.text


def test_multiple_iterations():
    """Test multiple iterations through the steps"""
    obj = DummyObject()

    # First iteration with key methods
    assert obj.step0_key() == "step0_key_result"
    assert obj.step1_key() == "step1_key_result"

    # Second iteration starting with set method
    assert obj.step0_set() == "step0_set_result"
    # Can't do step 1 with key method after set without redoing step 0 key
    assert obj.step1_key() == "step1_key_result"


def test_guide_decorator():
    """Test that the guide decorator properly wraps methods"""
    obj = DummyObject()

    # Check that the decorator preserves function metadata
    assert obj.step0_key.__name__ == "step0_key"
    assert obj.step0_getter.__name__ == "step0_getter"

    # Check that the decorator routes through the guide handler
    assert obj.step0_key() == "step0_key_result"
    assert obj.step0_getter() == "step0_get_result"



from zephyr_ml.labeling import utils
from zephyr_ml.labeling.data_labeler import DataLabeler
from zephyr_ml.labeling.labeling_functions import (
    brake_pad_presence, converter_replacement_presence, gearbox_replace_presence, total_power_loss)

LABELING_FUNCTIONS = [
    brake_pad_presence,
    converter_replacement_presence,
    gearbox_replace_presence,
    total_power_loss,
]

UTIL_FUNCTIONS = [
    utils.aggregate_by_column,
    utils.categorical_presence,
    utils.greater_than,
    utils.keyword_in_text,
    utils.merge_binary_labeling_functions,
    utils.total_duration,
]


def get_labeling_functions():
    functions = {}
    for function in LABELING_FUNCTIONS:
        name = function.__name__
        functions[name] = {"obj": function, "desc": function.__doc__.split("\n")[
            0]}

    return functions


def get_labeling_functions_map():
    functions = {}
    for function in LABELING_FUNCTIONS:
        name = function.__name__
        functions[name] = function
    return functions


def get_helper_functions():
    functions = {}
    for function in UTIL_FUNCTIONS:
        name = function.__name__
        functions[name] = function.__doc__.split("\n")[0]

    return functions


def get_util_functions():
    return get_helper_functions()



from unittest.mock import Mock

import featuretools as ft

from zephyr_ml.labeling import DataLabeler


class TestDataLabeler:

    @classmethod
    def setup_class(cls):
        cls.function = lambda x: x
        cls.es = Mock(autospec=ft.EntitySet)
        cls.subset = None
        cls.verbose = False

    def test_data_labeler(self):
        def function(x):
            return x

        DataLabeler(function)



import pytest

from zephyr_ml.metadata import DEFAULT_ES_KWARGS, DEFAULT_ES_TYPE_KWARGS, get_mapped_kwargs


def test_default_scada_mapped_kwargs():
    expected = {**DEFAULT_ES_KWARGS, 'scada': DEFAULT_ES_TYPE_KWARGS['scada']}
    actual = get_mapped_kwargs('scada')
    assert expected == actual


def test_default_pidata_mapped_kwargs():
    expected = {**DEFAULT_ES_KWARGS,
                'pidata': DEFAULT_ES_TYPE_KWARGS['pidata']}
    actual = get_mapped_kwargs('pidata')
    assert expected == actual


def test_new_kwargs_bad_es_type():
    error_text = "Unrecognized es_type argument: bad_es_type"
    with pytest.raises(ValueError, match=error_text):
        get_mapped_kwargs('bad_es_type')


def test_new_kwargs_bad_format():
    error_text = "new_kwargs must be dictionary mapping entity name to dictionary " + \
                 "with updated keyword arguments for EntitySet creation."
    bad_kwargs = ['list', 'of', 'args']
    with pytest.raises(ValueError, match=error_text):
        get_mapped_kwargs('pidata', bad_kwargs)


def test_new_kwargs_unexpected_entity():
    error_text = 'Unrecognized entity "unexpected" found in new keyword argument mapping.'
    bad_kwargs = {'unexpected': {}}
    with pytest.raises(ValueError, match=error_text):
        get_mapped_kwargs('pidata', bad_kwargs)


def test_new_kwargs_update_kwargs():
    updated_kwargs = {'turbines': {
        'index': 'new_turbine_index',
        'logical_types': {
            'COD_ELEMENT': 'integer',
            'extra_column': 'double'
        }
    }}

    expected = {
        **DEFAULT_ES_KWARGS,
        'pidata': DEFAULT_ES_TYPE_KWARGS['pidata'],
        'turbines': {
            'index': 'new_turbine_index',
            'logical_types': {
                'COD_ELEMENT': 'integer',
                'extra_column': 'double'
            }
        }
    }

    actual = get_mapped_kwargs('pidata', updated_kwargs)
    assert actual == expected



from inspect import getfullargspec

import composeml as cp


class DataLabeler:
    """Class that defines the prediction problem.
    This class supports the generation of `label_times` which
    is fundamental to the feature generation phase as well
    as specifying the target labels.
    Args:
        function (LabelingFunction):
            function that defines the labeling function, it should return a
            tuple of labeling function, the dataframe, and the name of the
            target entity.
    """

    def __init__(self, function):
        self.function = function

    def generate_label_times(self, es, num_samples=-1, subset=None,
                             column_map={}, verbose=False, **kwargs):
        """Searches the data to calculate label times.
          Args:
              es (featuretools.EntitySet):
                  Entityset to extract `label_times` from.
              num_samples (int):
                  Number of samples for each to return. Defaults to -1 which returns all possible
                  samples.
              subset (float or int):
                  Portion of the data to select for searching.
              verbose:
                  An indicator to the verbosity of searching.
              column_map:
                  Dictionary mapping column references in labeling function to actual column names.
                  See labeling function for columns referenced.
          Returns:
              composeml.LabelTimes:
                  Calculated labels with cutoff times.
        """
        labeling_function, df, meta = self.function(es, column_map)

        data = df
        if isinstance(subset, float) or isinstance(subset, int):
            data = data.sample(subset)

        target_entity_index = meta.get('target_entity_index')
        time_index = meta.get('time_index')
        thresh = kwargs.get('thresh') or meta.get('thresh')
        window_size = kwargs.get('window_size') or meta.get('window_size')
        label_maker = cp.LabelMaker(labeling_function=labeling_function,
                                    target_dataframe_name=target_entity_index,
                                    time_index=time_index,
                                    window_size=window_size)

        kwargs = {**meta, **kwargs}
        kwargs = {
            k: kwargs.get(k) for k in set(
                getfullargspec(
                    label_maker.search)[0]) if kwargs.get(k) is not None}
        label_times = label_maker.search(data.sort_values(time_index), num_samples,
                                         verbose=verbose, **kwargs)
        if thresh is not None:
            label_times = label_times.threshold(thresh)

        return label_times, meta



import featuretools as ft
import numpy as np
import pandas as pd

from zephyr_ml.labeling.utils import (
    aggregate_by_column, categorical_presence, denormalize, greater_than, keyword_in_text,
    merge_binary_labeling_functions, total_duration)


def test_aggregate_by_column():
    data = pd.DataFrame({
        'column': [1, 2, 3]
    })
    assert 2 == aggregate_by_column('column', np.mean)(data)


def test_merge_labeling_and_true():
    functions = [
        lambda df: True,
        lambda df: True
    ]
    assert 1 == merge_binary_labeling_functions(
        functions, and_connected=True)(pd.DataFrame())


def test_merge_labeling_and_false():
    functions = [
        lambda df: True,
        lambda df: False
    ]
    assert 0 == merge_binary_labeling_functions(
        functions, and_connected=True)(pd.DataFrame())


def test_merge_labeling_or_true():
    functions = [
        lambda df: False,
        lambda df: True
    ]
    assert 1 == merge_binary_labeling_functions(
        functions, and_connected=False)(pd.DataFrame())


def test_merge_labeling_or_false():
    functions = [
        lambda df: False,
        lambda df: False
    ]
    assert 0 == merge_binary_labeling_functions(
        functions, and_connected=False)(pd.DataFrame())


def test_categorical_presence_true():
    data = pd.DataFrame({
        'column': ['A', 'B', 'C']
    })
    function = categorical_presence('column', 'A')
    assert 1 == function(data)


def test_categorical_presence_false():
    data = pd.DataFrame({
        'column': ['A', 'B', 'C']
    })
    function = categorical_presence('column', 'D')
    assert 0 == function(data)


def test_keyword_in_text_true():
    data = pd.DataFrame({
        'A': ['this is a comment'],
        'B': ['this is a description']
    })
    function = keyword_in_text('description', columns=['A', 'B'])
    assert 1 == function(data)


def test_keyword_in_text_false():
    data = pd.DataFrame({
        'A': ['this is a comment'],
        'B': ['this is a description']
    })
    function = keyword_in_text('text', columns=['A', 'B'])
    assert 0 == function(data)


def test_keyword_in_unknown_column():
    data = pd.DataFrame({
        'A': ['this is a comment'],
        'B': ['this is a description']
    })
    function = keyword_in_text('text', columns=['A', 'B', 'C'])
    assert 0 == function(data)


def test_greater_than_true():
    data = pd.DataFrame({
        'A': [1, 2, 3, 4],
    })
    function = greater_than('A', 3)
    assert 1 == function(data)


def test_greater_than_false():
    data = pd.DataFrame({
        'A': [1, 2, 3, 4],
    })
    function = greater_than('A', 5)
    assert 0 == function(data)


def test_total_duration():
    data = pd.DataFrame({
        'Start': [pd.to_datetime('2010-01-01T10:00'), pd.to_datetime('2010-01-01T14:00')],
        'End': [pd.to_datetime('2010-01-01T11:00'), pd.to_datetime('2010-01-01T16:00')]
    })
    assert 10800 == total_duration('Start', 'End')(data)


def test_total_duration_nan():
    data = pd.DataFrame({
        'Start': [pd.to_datetime('2010-01-01T10:00'), np.nan],
        'End': [pd.to_datetime('2010-01-01T11:00'), pd.to_datetime('2010-01-01T18:00')]
    })
    assert 3600 == total_duration('Start', 'End')(data)


def test_denormalize():
    child_data = pd.DataFrame({
        'child_id': [0, 1, 2],
        'parent_id': ['a', 'a', 'b'],
        'child_value': [100, -5, 25]
    })
    parent_data = pd.DataFrame({
        'parent_id': ['a', 'b'],
        'parent_val': ['x', 'y']
    })

    es = ft.EntitySet(dataframes={'parent': (parent_data, 'parent_id'),
                                  'child': (child_data, 'child_id')},
                      relationships=[('parent', 'parent_id', 'child', 'parent_id')])

    expected = pd.DataFrame({
        'child_id': [0, 1, 2],
        'parent_id': ['a', 'a', 'b'],
        'parent_val': ['x', 'x', 'y'],
        'child_value': [100, -5, 25]
    })
    expected.ww.init()

    actual = denormalize(es, ['child', 'parent'])
    pd.testing.assert_frame_equal(expected, actual, check_like=True)

    actual2 = denormalize(es, ['parent', 'child'])
    pd.testing.assert_frame_equal(expected, actual2, check_like=True)



# -*- coding: utf-8 -*-

"""Top-level package for Zephyr."""

__author__ = 'MIT Data To AI Lab'
__email__ = 'dai-lab@mit.edu'
__version__ = '0.0.5.dev1'

import os

from zephyr_ml.core import Zephyr
from zephyr_ml.entityset import VALIDATE_DATA_FUNCTIONS, _create_entityset
from zephyr_ml.labeling import DataLabeler

MLBLOCKS_PRIMITIVES = os.path.join(
    os.path.dirname(__file__), "primitives", "jsons")
MLBLOCKS_PIPELINES = os.path.join(os.path.dirname(__file__), "pipelines")



import pandas as pd
import pytest

from zephyr_ml import _create_entityset
from zephyr_ml.feature_engineering import process_signals


@pytest.fixture
def base_dfs():
    alarms_df = pd.DataFrame({
        'COD_ELEMENT': [0, 0],
        'DAT_START': [pd.Timestamp('2022-01-01 00:00:00'), pd.Timestamp('2022-03-01 11:12:13')],
        'DAT_END': [pd.Timestamp('2022-01-01 13:00:00'), pd.Timestamp('2022-03-02 11:12:13')],
        'IND_DURATION': [0.5417, 1.0],
        'COD_ALARM': [12345, 98754],
        'COD_ALARM_INT': [12345, 98754],
        'DES_NAME': ['Alarm1', 'Alarm2'],
        'DES_TITLE': ['Description of alarm 1', 'Description of alarm 2'],
    })
    stoppages_df = pd.DataFrame({
        'COD_ELEMENT': [0, 0],
        'DAT_START': [pd.Timestamp('2022-01-01 00:00:00'), pd.Timestamp('2022-03-01 11:12:13')],
        'DAT_END': [pd.Timestamp('2022-01-08 11:07:17'), pd.Timestamp('2022-03-01 17:00:13')],
        'DES_WO_NAME': ['stoppage name 1', 'stoppage name 2'],
        'DES_COMMENTS': ['description of stoppage 1', 'description of stoppage 2'],
        'COD_WO': [12345, 67890],
        'IND_DURATION': [7.4642, 0.2417],
        'IND_LOST_GEN': [45678.0, 123.0],
        'COD_ALARM': [12345, 12345],
        'COD_CAUSE': [32, 48],
        'COD_INCIDENCE': [987654, 123450],
        'COD_ORIGIN': [6, 23],
        'COD_STATUS': ['STOP', 'PAUSE'],
        'COD_CODE': ['ABC', 'XYZ'],
        'DES_DESCRIPTION': ['Description 1', 'Description 2']
    })
    notifications_df = pd.DataFrame({
        'COD_ELEMENT': [0, 0],
        'COD_ORDER': [12345, 67890],
        'IND_QUANTITY': [1, -20],
        'COD_MATERIAL_SAP': [36052411, 67890],
        'DAT_POSTING': [pd.Timestamp('2022-01-01 00:00:00'), pd.Timestamp('2022-03-01 00:00:00')],
        'COD_MAT_DOC': [77889900, 12345690],
        'DES_MEDIUM': ['Description of notification 1', 'Description of notification 2'],
        'COD_NOTIF': [567890123, 32109877],
        'DAT_MALF_START': [pd.Timestamp('2021-12-25 18:07:10'),
                           pd.Timestamp('2022-02-28 06:04:00')],
        'DAT_MALF_END': [pd.Timestamp('2022-01-08 11:07:17'), pd.Timestamp('2022-03-01 17:00:13')],
        'IND_BREAKDOWN_DUR': [14.1378, 2.4792],
        'FUNCT_LOC_DES': ['location description 1', 'location description 2'],
        'COD_ALARM': [12345, 12345],
        'DES_ALARM': ['Alarm description', 'Alarm description'],
    })
    work_orders_df = pd.DataFrame({
        'COD_ELEMENT': [0, 0],
        'COD_ORDER': [12345, 67890],
        'DAT_BASIC_START': [pd.Timestamp('2022-01-01 00:00:00'),
                            pd.Timestamp('2022-03-01 00:00:00')],
        'DAT_BASIC_END': [pd.Timestamp('2022-01-09 00:00:00'),
                          pd.Timestamp('2022-03-02 00:00:00')],
        'COD_EQUIPMENT': [98765, 98765],
        'COD_MAINT_PLANT': ['ABC', 'ABC'],
        'COD_MAINT_ACT_TYPE': ['XYZ', 'XYZ'],
        'COD_CREATED_BY': ['A1234', 'B6789'],
        'COD_ORDER_TYPE': ['A', 'B'],
        'DAT_REFERENCE': [pd.Timestamp('2022-01-01 00:00:00'),
                          pd.Timestamp('2022-03-01 00:00:00')],
        'DAT_CREATED_ON': [pd.Timestamp('2022-03-01 00:00:00'),
                           pd.Timestamp('2022-04-18 00:00:00')],
        'DAT_VALID_END': [pd.NaT, pd.NaT],
        'DAT_VALID_START': [pd.NaT, pd.NaT],
        'COD_SYSTEM_STAT': ['ABC XYZ', 'LMN OPQ'],
        'DES_LONG': ['description of work order', 'description of work order'],
        'COD_FUNCT_LOC': ['!12345', '?09876'],
        'COD_NOTIF_OBJ': ['00112233', '00998877'],
        'COD_MAINT_ITEM': ['', '019283'],
        'DES_MEDIUM': ['short description', 'short description'],
        'DES_FUNCT_LOC': ['XYZ1234', 'ABC9876'],
    })
    turbines_df = pd.DataFrame({
        'COD_ELEMENT': [0],
        'TURBINE_PI_ID': ['TA00'],
        'TURBINE_LOCAL_ID': ['A0'],
        'TURBINE_SAP_COD': ['LOC000'],
        'DES_CORE_ELEMENT': ['T00'],
        'SITE': ['LOCATION'],
        'DES_CORE_PLANT': ['LOC'],
        'COD_PLANT_SAP': ['ABC'],
        'PI_COLLECTOR_SITE_NAME': ['LOC0'],
        'PI_LOCAL_SITE_NAME': ['LOC0']
    })
    return {
        'alarms': alarms_df,
        'stoppages': stoppages_df,
        'notifications': notifications_df,
        'work_orders': work_orders_df,
        'turbines': turbines_df
    }


@pytest.fixture
def pidata_dfs(base_dfs):
    pidata_df = pd.DataFrame({
        'time': [pd.Timestamp('2022-01-02 13:21:01'), pd.Timestamp('2022-03-08 13:21:01')],
        'COD_ELEMENT': [0, 0],
        'val1': [9872.0, 559.0],
        'val2': [10.0, -7.0]
    })
    return {**base_dfs, 'pidata': pidata_df}


@pytest.fixture
def scada_dfs(base_dfs):
    scada_df = pd.DataFrame({
        'TIMESTAMP': [pd.Timestamp('2022-01-02 13:21:01'), pd.Timestamp('2022-03-08 13:21:01')],
        'COD_ELEMENT': [0, 0],
        'val1': [1002.0, 56.8],
        'val2': [-98.7, 1004.2]
    })
    return {**base_dfs, 'scada': scada_df}


@pytest.fixture
def pidata_es(pidata_dfs):
    return _create_entityset(pidata_dfs, "pidata")


@pytest.fixture
def scada_es(scada_dfs):
    return _create_entityset(scada_dfs, "scada")


@pytest.fixture
def transformations():
    return [{
        "name": "fft",
        "primitive": "sigpro.transformations.amplitude.identity.identity"
    }]


@pytest.fixture
def aggregations():
    return [{
        "name": "mean",
        "primitive": "sigpro.aggregations.amplitude.statistical.mean"
    }]


def test_process_signals_pidata(pidata_es, transformations, aggregations):
    signal_dataframe_name = 'pidata'
    signal_column = 'val1'
    window_size = '1m'
    replace_dataframe = False
    before = pidata_es['pidata'].copy()

    process_signals(pidata_es, signal_dataframe_name, signal_column, transformations, aggregations,
                    window_size, replace_dataframe)

    processed = pidata_es['pidata_processed'].copy()
    after = pidata_es['pidata'].copy()

    expected = pd.DataFrame({
        "_index": [0, 1, 2],
        "COD_ELEMENT": [0, 0, 0],
        "time": [
            pd.Timestamp('2022-01-31'),
            pd.Timestamp('2022-02-28'),
            pd.Timestamp('2022-03-31')
        ],
        "fft.mean.mean_value": [9872, None, 559]
    })
    expected['COD_ELEMENT'] = expected['COD_ELEMENT'].astype('category')
    expected['fft.mean.mean_value'] = expected['fft.mean.mean_value'].astype(
        'float64')
    processed['fft.mean.mean_value'] = processed['fft.mean.mean_value'].astype(
        'float64')

    assert pidata_es['pidata_processed'].shape[0] == 3
    assert pidata_es['pidata_processed'].shape[1] == 4

    pd.testing.assert_frame_equal(before, after)
    pd.testing.assert_frame_equal(processed, expected)


def test_process_signals_pidata_replace(
        pidata_es, transformations, aggregations):
    signal_dataframe_name = 'pidata'
    signal_column = 'val1'
    window_size = '1m'
    replace_dataframe = True

    process_signals(pidata_es, signal_dataframe_name, signal_column, transformations, aggregations,
                    window_size, replace_dataframe)

    processed = pidata_es['pidata'].copy()

    expected = pd.DataFrame({
        "_index": [0, 1, 2],
        "COD_ELEMENT": [0, 0, 0],
        "time": [
            pd.Timestamp('2022-01-31'),
            pd.Timestamp('2022-02-28'),
            pd.Timestamp('2022-03-31')
        ],
        "fft.mean.mean_value": [9872, None, 559]
    })
    expected['COD_ELEMENT'] = expected['COD_ELEMENT'].astype('category')
    expected['fft.mean.mean_value'] = expected['fft.mean.mean_value'].astype(
        'float64')
    processed['fft.mean.mean_value'] = processed['fft.mean.mean_value'].astype(
        'float64')

    assert pidata_es['pidata'].shape[0] == 3
    assert pidata_es['pidata'].shape[1] == 4

    pd.testing.assert_frame_equal(processed, expected)
    assert 'pidata_processed' not in list(pidata_es.dataframe_dict.keys())


def test_process_signals_scada(scada_es, transformations, aggregations):
    signal_dataframe_name = 'scada'
    signal_column = 'val1'
    window_size = '1m'
    replace_dataframe = False
    before = scada_es['scada'].copy()

    process_signals(scada_es, signal_dataframe_name, signal_column, transformations, aggregations,
                    window_size, replace_dataframe)

    expected = pd.DataFrame({
        "_index": [0, 1, 2],
        "COD_ELEMENT": [0, 0, 0],
        "TIMESTAMP": [
            pd.Timestamp('2022-01-31'),
            pd.Timestamp('2022-02-28'),
            pd.Timestamp('2022-03-31')
        ],
        "fft.mean.mean_value": [1002, None, 56.8]
    })
    expected['COD_ELEMENT'] = expected['COD_ELEMENT'].astype('category')
    expected['fft.mean.mean_value'] = expected['fft.mean.mean_value'].astype(
        'float64')
    after = scada_es['scada'].copy()

    assert scada_es['scada_processed'].shape[0] == 3
    assert scada_es['scada_processed'].shape[1] == 4

    pd.testing.assert_frame_equal(before, after)
    pd.testing.assert_frame_equal(scada_es['scada_processed'], expected)


def test_process_signals_scada_replace(
        scada_es, transformations, aggregations):
    signal_dataframe_name = 'scada'
    signal_column = 'val1'
    window_size = '1m'
    replace_dataframe = True

    process_signals(scada_es, signal_dataframe_name, signal_column, transformations, aggregations,
                    window_size, replace_dataframe)

    expected = pd.DataFrame({
        "_index": [0, 1, 2],
        "COD_ELEMENT": [0, 0, 0],
        "TIMESTAMP": [
            pd.Timestamp('2022-01-31'),
            pd.Timestamp('2022-02-28'),
            pd.Timestamp('2022-03-31')
        ],
        "fft.mean.mean_value": [1002, None, 56.8]
    })
    expected['COD_ELEMENT'] = expected['COD_ELEMENT'].astype('category')
    expected['fft.mean.mean_value'] = expected['fft.mean.mean_value'].astype(
        'float64')

    assert scada_es['scada'].shape[0] == 3
    assert scada_es['scada'].shape[1] == 4

    pd.testing.assert_frame_equal(scada_es['scada'], expected)
    assert 'scada_processed' not in list(scada_es.dataframe_dict.keys())



import numpy as np
import pandas as pd
from mlblocks import MLBlock

from zephyr_ml.core import DEFAULT_METRICS, Zephyr


class TestZephyr:

    @staticmethod
    def base_dfs():
        alarms_df = pd.DataFrame({
            'COD_ELEMENT': [0, 0],
            'DAT_START': [pd.Timestamp('2022-01-01 00:00:00'),
                          pd.Timestamp('2022-03-01 11:12:13')],
            'DAT_END': [pd.Timestamp('2022-01-01 13:00:00'),
                        pd.Timestamp('2022-03-02 11:12:13')],
            'IND_DURATION': [0.5417, 1.0],
            'COD_ALARM': [12345, 98754],
            'COD_ALARM_INT': [12345, 98754],
            'DES_NAME': ['Alarm1', 'Alarm2'],
            'DES_TITLE': ['Description of alarm 1', 'Description of alarm 2'],
        })
        stoppages_df = pd.DataFrame({
            'COD_ELEMENT': [0, 0],
            'DAT_START': [pd.Timestamp('2022-01-01 00:00:00'),
                          pd.Timestamp('2022-03-01 11:12:13')],
            'DAT_END': [pd.Timestamp('2022-01-08 11:07:17'),
                        pd.Timestamp('2022-03-01 17:00:13')],
            'DES_WO_NAME': ['stoppage name 1', 'stoppage name 2'],
            'DES_COMMENTS': ['description of stoppage 1', 'description of stoppage 2'],
            'COD_WO': [12345, 67890],
            'IND_DURATION': [7.4642, 0.2417],
            'IND_LOST_GEN': [45678.0, 123.0],
            'COD_ALARM': [12345, 12345],
            'COD_CAUSE': [32, 48],
            'COD_INCIDENCE': [987654, 123450],
            'COD_ORIGIN': [6, 23],
            'COD_STATUS': ['STOP', 'PAUSE'],
            'COD_CODE': ['ABC', 'XYZ'],
            'DES_DESCRIPTION': ['Description 1', 'Description 2']
        })
        notifications_df = pd.DataFrame({
            'COD_ELEMENT': [0, 0],
            'COD_ORDER': [12345, 67890],
            'IND_QUANTITY': [1, -20],
            'COD_MATERIAL_SAP': [36052411, 67890],
            'DAT_POSTING': [pd.Timestamp('2022-01-01 00:00:00'),
                            pd.Timestamp('2022-03-01 00:00:00')],
            'COD_MAT_DOC': [77889900, 12345690],
            'DES_MEDIUM': ['Description of notification 1', 'Description of notification 2'],
            'COD_NOTIF': [567890123, 32109877],
            'DAT_MALF_START': [pd.Timestamp('2021-12-25 18:07:10'),
                               pd.Timestamp('2022-02-28 06:04:00')],
            'DAT_MALF_END': [pd.Timestamp('2022-01-08 11:07:17'),
                             pd.Timestamp('2022-03-01 17:00:13')],
            'IND_BREAKDOWN_DUR': [14.1378, 2.4792],
            'FUNCT_LOC_DES': ['location description 1', 'location description 2'],
            'COD_ALARM': [12345, 12345],
            'DES_ALARM': ['Alarm description', 'Alarm description'],
        })
        work_orders_df = pd.DataFrame({
            'COD_ELEMENT': [0, 0],
            'COD_ORDER': [12345, 67890],
            'DAT_BASIC_START': [pd.Timestamp('2022-01-01 00:00:00'),
                                pd.Timestamp('2022-03-01 00:00:00')],
            'DAT_BASIC_END': [pd.Timestamp('2022-01-09 00:00:00'),
                              pd.Timestamp('2022-03-02 00:00:00')],
            'COD_EQUIPMENT': [98765, 98765],
            'COD_MAINT_PLANT': ['ABC', 'ABC'],
            'COD_MAINT_ACT_TYPE': ['XYZ', 'XYZ'],
            'COD_CREATED_BY': ['A1234', 'B6789'],
            'COD_ORDER_TYPE': ['A', 'B'],
            'DAT_REFERENCE': [pd.Timestamp('2022-01-01 00:00:00'),
                              pd.Timestamp('2022-03-01 00:00:00')],
            'DAT_CREATED_ON': [pd.Timestamp('2022-03-01 00:00:00'),
                               pd.Timestamp('2022-04-18 00:00:00')],
            'DAT_VALID_END': [pd.NaT, pd.NaT],
            'DAT_VALID_START': [pd.NaT, pd.NaT],
            'COD_SYSTEM_STAT': ['ABC XYZ', 'LMN OPQ'],
            'DES_LONG': ['description of work order', 'description of work order'],
            'COD_FUNCT_LOC': ['!12345', '?09876'],
            'COD_NOTIF_OBJ': ['00112233', '00998877'],
            'COD_MAINT_ITEM': ['', '019283'],
            'DES_MEDIUM': ['short description', 'short description'],
            'DES_FUNCT_LOC': ['XYZ1234', 'ABC9876'],
        })
        turbines_df = pd.DataFrame({
            'COD_ELEMENT': [0],
            'TURBINE_PI_ID': ['TA00'],
            'TURBINE_LOCAL_ID': ['A0'],
            'TURBINE_SAP_COD': ['LOC000'],
            'DES_CORE_ELEMENT': ['T00'],
            'SITE': ['LOCATION'],
            'DES_CORE_PLANT': ['LOC'],
            'COD_PLANT_SAP': ['ABC'],
            'PI_COLLECTOR_SITE_NAME': ['LOC0'],
            'PI_LOCAL_SITE_NAME': ['LOC0']
        })
        pidata_df = pd.DataFrame({
            'time': [pd.Timestamp('2022-01-02 13:21:01'),
                     pd.Timestamp('2022-03-08 13:21:01')],
            'COD_ELEMENT': [0, 0],
            'val1': [9872.0, 559.0],
            'val2': [10.0, -7.0]
        })
        return {
            'alarms': alarms_df,
            'stoppages': stoppages_df,
            'notifications': notifications_df,
            'work_orders': work_orders_df,
            'turbines': turbines_df,
            "pidata": pidata_df
        }

    def base_train_test_split(self):
        X_train = pd.DataFrame({
            'feature 1': np.random.random(300),
            'feature 2': [0] * 150 + [1] * 150,
        })
        y_train = X_train['feature 2'].to_list()
        X_test = pd.DataFrame({
            'feature 1': np.random.random((100)),
            'feature 2': [0] * 25 + [1] * 50 + [0] * 25,
        })
        y_test = X_test['feature 2'].to_list()
        return X_train, X_test, y_train, y_test

    @classmethod
    def setup_class(cls):
        cls.train = pd.DataFrame({
            'feature 1': np.random.random(300),
            'feature 2': [0] * 150 + [1] * 150,
        })
        cls.train_y = cls.train['feature 2'].to_list()
        cls.test = pd.DataFrame({
            'feature 1': np.random.random((100)),
            'feature 2': [0] * 25 + [1] * 50 + [0] * 25,
        })
        cls.test_y = cls.test['feature 2'].to_list()
        cls.random = pd.DataFrame({
            'feature 1': list(range(100)),
            'feature 2': np.random.random(100),
            'feature 3': np.random.random(100),
        })
        cls.random_y = [1 if x > 0.5 else 0 for x in np.random.random(100)]
        cls.kwargs = {
            "generate_entityset": {
                "dfs": TestZephyr.base_dfs(),
                "es_type": "pidata"},
            "generate_label_times": {
                "labeling_fn": "brake_pad_presence",
                "num_samples": 10,
                "gap": "20d"},
            "generate_feature_matrix": {
                "target_dataframe_name": "turbines",
                "cutoff_time_in_index": True,
                "agg_primitives": [
                    "count",
                    "sum",
                    "max"],
                "verbose": True},
            "generate_train_test_split": {},
            "fit_pipeline": {},
            "evaluate": {}}

    def test_initialize_class(self):
        _ = Zephyr()

    def test_generate_entityset(self):
        zephyr = Zephyr()
        zephyr.generate_entityset(
            **self.__class__.kwargs["generate_entityset"])
        es = zephyr.get_entityset()
        assert es is not None
        assert es.id == 'pidata'

    def test_generate_label_times(self):
        zephyr = Zephyr()
        zephyr.generate_entityset(
            **self.__class__.kwargs["generate_entityset"])
        zephyr.generate_label_times(
            **self.__class__.kwargs["generate_label_times"])
        label_times = zephyr.get_label_times(visualize=False)
        assert label_times is not None

    def test_generate_feature_matrix_and_labels(self):
        zephyr = Zephyr()
        zephyr.generate_entityset(
            **self.__class__.kwargs["generate_entityset"])
        zephyr.generate_label_times(
            **self.__class__.kwargs["generate_label_times"])
        zephyr.generate_feature_matrix(
            **self.__class__.kwargs["generate_feature_matrix"])
        feature_matrix, label_col_name, features = zephyr.get_feature_matrix()
        assert feature_matrix is not None
        assert label_col_name in feature_matrix.columns
        assert features is not None

    def test_generate_train_test_split(self):
        zephyr = Zephyr()
        zephyr.generate_entityset(
            **self.__class__.kwargs["generate_entityset"])
        zephyr.generate_label_times(
            **self.__class__.kwargs["generate_label_times"])
        zephyr.generate_feature_matrix(
            **self.__class__.kwargs["generate_feature_matrix"])
        zephyr.generate_train_test_split(
            **self.__class__.kwargs["generate_train_test_split"])
        train_test_split = zephyr.get_train_test_split()
        assert train_test_split is not None
        X_train, X_test, y_train, y_test = train_test_split
        assert isinstance(X_train, pd.DataFrame)
        assert isinstance(X_test, pd.DataFrame)
        assert isinstance(y_train, pd.Series)
        assert isinstance(y_test, pd.Series)

    def test_set_train_test_split(self):
        zephyr = Zephyr()
        zephyr.set_train_test_split(*self.base_train_test_split())
        train_test_split = zephyr.get_train_test_split()
        assert train_test_split is not None
        X_train, X_test, y_train, y_test = train_test_split
        assert isinstance(X_train, pd.DataFrame)
        assert isinstance(X_test, pd.DataFrame)
        assert isinstance(y_train, list)
        assert isinstance(y_test, list)

    def test_fit_pipeline_no_visual(self):
        zephyr = Zephyr()
        zephyr.set_train_test_split(*self.base_train_test_split())
        output = zephyr.fit_pipeline(**self.__class__.kwargs["fit_pipeline"])
        assert output is None
        pipeline = zephyr.get_fitted_pipeline()
        assert pipeline is not None

    def test_fit_pipeline_visual(self):
        zephyr = Zephyr()
        zephyr.set_train_test_split(*self.base_train_test_split())
        output = zephyr.fit_pipeline(
            visual=True, **self.__class__.kwargs["fit_pipeline"])
        assert isinstance(output, dict)
        assert list(output.keys()) == ['threshold', 'scores']
        pipeline = zephyr.get_fitted_pipeline()
        assert pipeline is not None

    def test_predict_no_visual(self):
        zephyr = Zephyr()
        zephyr.set_train_test_split(*self.base_train_test_split())
        zephyr.fit_pipeline(**self.__class__.kwargs["fit_pipeline"])
        predicted = zephyr.predict()
        _, _, _, test_y = self.base_train_test_split()
        assert predicted == test_y

    def test_predict_visual(self):
        zephyr = Zephyr()
        zephyr.set_train_test_split(*self.base_train_test_split())
        zephyr.fit_pipeline(**self.__class__.kwargs["fit_pipeline"])
        predicted, output = zephyr.predict(visual=True)
        assert isinstance(predicted, list)
        assert len(predicted) == len(self.test_y)
        assert isinstance(output, dict)
        assert list(output.keys()) == ['threshold', 'scores']

    def test_evaluate(self):
        zephyr = Zephyr()
        zephyr.set_train_test_split(*self.base_train_test_split())
        zephyr.fit_pipeline(**self.__class__.kwargs["fit_pipeline"])
        scores = zephyr.evaluate(metrics=[
            "sklearn.metrics.accuracy_score",
            "sklearn.metrics.precision_score",
            "sklearn.metrics.f1_score",
            "sklearn.metrics.recall_score"
        ])
        assert isinstance(scores, dict)
        assert all(metric in scores for metric in [
            "sklearn.metrics.accuracy_score",
            "sklearn.metrics.precision_score",
            "sklearn.metrics.f1_score",
            "sklearn.metrics.recall_score"
        ])

    def test_get_entityset_types(self):
        zephyr = Zephyr()
        entityset_types = zephyr.GET_ENTITYSET_TYPES()
        assert isinstance(entityset_types, dict)
        assert "pidata" in entityset_types
        assert "scada" in entityset_types
        assert "vibrations" in entityset_types
        for es_type, info in entityset_types.items():
            assert isinstance(info, dict)
            assert "obj" in info
            assert "desc" in info
            assert isinstance(info["obj"], str)
            assert isinstance(info["desc"], str)

    def test_get_labeling_functions(self):
        zephyr = Zephyr()
        labeling_functions = zephyr.GET_LABELING_FUNCTIONS()
        assert isinstance(labeling_functions, dict)
        assert "brake_pad_presence" in labeling_functions
        for func_name, info in labeling_functions.items():
            assert isinstance(info, dict)
            assert "obj" in info
            assert "desc" in info
            assert callable(info["obj"])
            assert isinstance(info["desc"], str)

    def test_get_evaluation_metrics(self):
        zephyr = Zephyr()
        evaluation_metrics = zephyr.GET_EVALUATION_METRICS()
        assert isinstance(evaluation_metrics, dict)
        expected_metrics = DEFAULT_METRICS
        for metric in expected_metrics:
            assert metric in evaluation_metrics
        for metric_name, info in evaluation_metrics.items():
            assert isinstance(info, dict)
            assert "obj" in info
            assert "desc" in info
            assert isinstance(info["obj"], MLBlock)
            assert hasattr(info["obj"], "metadata")
            assert isinstance(info["desc"], str)



"""Zephyr Core module.

This module defines the Zephyr Class, which is responsible for the
model training and inference with the underlying MLBlocks pipelines.
"""

import json
import logging
import os
import pickle
from functools import partial
from typing import List, Union

import numpy as np
import pandas as pd
from mlblocks import MLPipeline
from sklearn import metrics

LOGGER = logging.getLogger(__name__)


_REGRESSION_METRICS = {
    "mae": metrics.mean_absolute_error,
    "mse": metrics.mean_squared_error,
    "r2": metrics.r2_score,
}

_CLASSIFICATION_METRICS = {
    "accuracy": metrics.accuracy_score,
    "f1": metrics.f1_score,
    "recall": metrics.recall_score,
    "precision": metrics.precision_score,
}

METRICS = _CLASSIFICATION_METRICS


class Zephyr:
    """Zephyr Class.

    The Zephyr Class provides the main machine learning pipeline functionalities
    of Zephyr and is responsible for the interaction with the underlying
    MLBlocks pipelines.

    Args:
        pipeline (str, dict or MLPipeline):
            Pipeline to use. It can be passed as:
                * An ``str`` with a path to a JSON file.
                * An ``str`` with the name of a registered pipeline.
                * An ``MLPipeline`` instance.
                * A ``dict`` with an ``MLPipeline`` specification.
        hyperparameters (dict):
            Additional hyperparameters to set to the Pipeline.
    """

    DEFAULT_PIPELINE = "xgb_classifier"

    def _get_mlpipeline(self):
        pipeline = self._pipeline
        if isinstance(pipeline, str) and os.path.isfile(pipeline):
            with open(pipeline) as json_file:
                pipeline = json.load(json_file)

        mlpipeline = MLPipeline(pipeline)
        if self._hyperparameters:
            mlpipeline.set_hyperparameters(self._hyperparameters)

        return mlpipeline

    def __init__(
        self,
        pipeline: Union[str, dict, MLPipeline] = None,
        hyperparameters: dict = None,
    ):
        self._pipeline = pipeline or self.DEFAULT_PIPELINE
        self._hyperparameters = hyperparameters
        self._mlpipeline = self._get_mlpipeline()
        self._fitted = False

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self._pipeline == other._pipeline
            and self._hyperparameters == other._hyperparameters
            and self._fitted == other._fitted
        )

    def _get_outputs_spec(self, default=True):
        outputs_spec = ["default"] if default else []

        try:
            visual_names = self._mlpipeline.get_output_names("visual")
            outputs_spec.append("visual")
        except ValueError:
            visual_names = []

        return outputs_spec, visual_names

    def fit(
        self,
        X: pd.DataFrame,
        y: Union[pd.Series, np.ndarray],
        visual: bool = False,
        **kwargs
    ):
        """Fit the pipeline to the given data.

        Args:
            X (DataFrame):
                Input data, passed as a ``pandas.DataFrame`` containing
                the feature matrix.
            y (Series or ndarray):
                Target data, passed as a ``pandas.Series`` or ``numpy.ndarray``
                containing the target values.
            visual (bool):
                If ``True``, capture the ``visual`` named output from the
                ``MLPipeline`` and return it as an output.
        """
        if not self._fitted:
            self._mlpipeline = self._get_mlpipeline()

        if visual:
            outputs_spec, visual_names = self._get_outputs_spec(False)
        else:
            outputs_spec = None

        outputs = self._mlpipeline.fit(X, y, output_=outputs_spec, **kwargs)
        self._fitted = True

        if visual and outputs is not None:
            return dict(zip(visual_names, outputs))

    def predict(self, X: pd.DataFrame, visual: bool = False,
                **kwargs) -> pd.Series:
        """Predict the pipeline to the given data.

        Args:
            X (DataFrame):

                Input data, passed as a ``pandas.DataFrame`` containing
                the feature matrix.
        visual (bool):
                If ``True``, capture the ``visual`` named output from the
                ``MLPipeline`` and return it as an output.

        Returns:
            Series or ndarray:
                Predictions to the input data.
        """
        if visual:
            outputs_spec, visual_names = self._get_outputs_spec()
        else:
            outputs_spec = "default"

        outputs = self._mlpipeline.predict(X, output_=outputs_spec, **kwargs)

        if visual and visual_names:
            prediction = outputs[0]
            return prediction, dict(
                zip(visual_names, outputs[-len(visual_names):]))

        return outputs

    def fit_predict(
        self, X: pd.DataFrame, y: Union[pd.Series, np.ndarray], **kwargs
    ) -> pd.Series:
        """Fit the pipeline to the data and then predict targets.

        This method is functionally equivalent to calling ``fit(X, y)``
        and later on ``predict(X)`` but with the difference that
        here the ``MLPipeline`` is called only once, using its ``fit``
        method, and the output is directly captured without having
        to execute the whole pipeline again during the ``predict`` phase.

        Args:
            X (DataFrame):
                Input data, passed as a ``pandas.DataFrame`` containing
                the feature matrix.
            y (Series or ndarray):
                Target data, passed as a ``pandas.Series`` or ``numpy.ndarray``
                containing the target values.

        Returns:
            Series or ndarray:
                Predictions to the input data.
        """
        if not self._fitted:
            self._mlpipeline = self._get_mlpipeline()

        result = self._mlpipeline.fit(X, y, output_="default", **kwargs)
        self._fitted = True

        return result

    def evaluate(
        self,
        X: pd.DataFrame,
        y: Union[pd.Series, np.ndarray],
        fit: bool = False,
        train_X: pd.DataFrame = None,
        train_y: Union[pd.Series, np.ndarray] = None,
        metrics: List[str] = METRICS,
    ) -> pd.Series:
        """Evaluate the performance of the pipeline.

        Args:
            X (DataFrame):
                Input data, passed as a ``pandas.DataFrame`` containing
                the feature matrix.
            y (Series or ndarray):
                Target data, passed as a ``pandas.Series`` or ``numpy.ndarray``
                containing the target values.
            fit (bool):
                Whether to fit the pipeline before evaluating it.
                Defaults to ``False``.
            train_X (DataFrame):
                Training data, passed as a ``pandas.DataFrame`` containing
                the feature matrix.
                If not given, the pipeline is fitted on ``X``.
            train_y (Series or ndarray):
                Target data used for training, passed as a ``pandas.Series`` or
                ``numpy.ndarray`` containing the target values.
            metrics (list):
                List of metrics to used passed as a list of strings.
                If not given, it defaults to all the metrics.

        Returns:
            Series:
                ``pandas.Series`` containing one element for each
                metric applied, with the metric name as index.
        """
        if not fit:
            method = self._mlpipeline.predict
        else:
            if not self._fitted:
                mlpipeline = self._get_mlpipeline()
            else:
                mlpipeline = self._mlpipeline

            if train_X is not None and train_y is not None:
                # fit first and then predict
                mlpipeline.fit(train_X, train_y)
                method = mlpipeline.predict
            else:
                # fit and predict at once
                method = partial(mlpipeline.fit, y=y, output_="default")

        result = method(X)

        scores = {metric: METRICS[metric](y, result) for metric in metrics}

        return pd.Series(scores)

    def save(self, path: str):
        """Save this object using pickle.

        Args:
            path (str):
                Path to the file where the serialization of
                this object will be stored.
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "wb") as pickle_file:
            pickle.dump(self, pickle_file)

    @classmethod
    def load(cls, path: str):
        """Load an Zephyr instance from a pickle file.

        Args:
            path (str):
                Path to the file where the instance has been
                previously serialized.

        Returns:
            Orion

        Raises:
            ValueError:
                If the serialized object is not a Zephyr instance.
        """
        with open(path, "rb") as pickle_file:
            zephyr = pickle.load(pickle_file)
            if not isinstance(zephyr, cls):
                raise ValueError("Serialized object is not a Zephyr instance")

            return zephyr



import pandas as pd
import pytest

from zephyr_ml import _create_entityset


@pytest.fixture
def base_dfs():
    alarms_df = pd.DataFrame({
        'COD_ELEMENT': [0, 0],
        'DAT_START': [pd.Timestamp('2022-01-01 00:00:00'), pd.Timestamp('2022-03-01 11:12:13')],
        'DAT_END': [pd.Timestamp('2022-01-01 13:00:00'), pd.Timestamp('2022-03-02 11:12:13')],
        'IND_DURATION': [0.5417, 1.0],
        'COD_ALARM': [12345, 98754],
        'COD_ALARM_INT': [12345, 98754],
        'DES_NAME': ['Alarm1', 'Alarm2'],
        'DES_TITLE': ['Description of alarm 1', 'Description of alarm 2'],
    })
    stoppages_df = pd.DataFrame({
        'COD_ELEMENT': [0, 0],
        'DAT_START': [pd.Timestamp('2022-01-01 00:00:00'), pd.Timestamp('2022-03-01 11:12:13')],
        'DAT_END': [pd.Timestamp('2022-01-08 11:07:17'), pd.Timestamp('2022-03-01 17:00:13')],
        'DES_WO_NAME': ['stoppage name 1', 'stoppage name 2'],
        'DES_COMMENTS': ['description of stoppage 1', 'description of stoppage 2'],
        'COD_WO': [12345, 67890],
        'IND_DURATION': [7.4642, 0.2417],
        'IND_LOST_GEN': [45678.0, 123.0],
        'COD_ALARM': [12345, 12345],
        'COD_CAUSE': [32, 48],
        'COD_INCIDENCE': [987654, 123450],
        'COD_ORIGIN': [6, 23],
        'COD_STATUS': ['STOP', 'PAUSE'],
        'COD_CODE': ['ABC', 'XYZ'],
        'DES_DESCRIPTION': ['Description 1', 'Description 2']
    })
    notifications_df = pd.DataFrame({
        'COD_ELEMENT': [0, 0],
        'COD_ORDER': [12345, 67890],
        'IND_QUANTITY': [1, -20],
        'COD_MATERIAL_SAP': [36052411, 67890],
        'DAT_POSTING': [pd.Timestamp('2022-01-01 00:00:00'), pd.Timestamp('2022-03-01 00:00:00')],
        'COD_MAT_DOC': [77889900, 12345690],
        'DES_MEDIUM': ['Description of notification 1', 'Description of notification 2'],
        'COD_NOTIF': [567890123, 32109877],
        'DAT_MALF_START': [pd.Timestamp('2021-12-25 18:07:10'),
                           pd.Timestamp('2022-02-28 06:04:00')],
        'DAT_MALF_END': [pd.Timestamp('2022-01-08 11:07:17'), pd.Timestamp('2022-03-01 17:00:13')],
        'IND_BREAKDOWN_DUR': [14.1378, 2.4792],
        'FUNCT_LOC_DES': ['location description 1', 'location description 2'],
        'COD_ALARM': [12345, 12345],
        'DES_ALARM': ['Alarm description', 'Alarm description'],
    })
    work_orders_df = pd.DataFrame({
        'COD_ELEMENT': [0, 0],
        'COD_ORDER': [12345, 67890],
        'DAT_BASIC_START': [pd.Timestamp('2022-01-01 00:00:00'),
                            pd.Timestamp('2022-03-01 00:00:00')],
        'DAT_BASIC_END': [pd.Timestamp('2022-01-09 00:00:00'),
                          pd.Timestamp('2022-03-02 00:00:00')],
        'COD_EQUIPMENT': [98765, 98765],
        'COD_MAINT_PLANT': ['ABC', 'ABC'],
        'COD_MAINT_ACT_TYPE': ['XYZ', 'XYZ'],
        'COD_CREATED_BY': ['A1234', 'B6789'],
        'COD_ORDER_TYPE': ['A', 'B'],
        'DAT_REFERENCE': [pd.Timestamp('2022-01-01 00:00:00'),
                          pd.Timestamp('2022-03-01 00:00:00')],
        'DAT_CREATED_ON': [pd.Timestamp('2022-03-01 00:00:00'),
                           pd.Timestamp('2022-04-18 00:00:00')],
        'DAT_VALID_END': [pd.NaT, pd.NaT],
        'DAT_VALID_START': [pd.NaT, pd.NaT],
        'COD_SYSTEM_STAT': ['ABC XYZ', 'LMN OPQ'],
        'DES_LONG': ['description of work order', 'description of work order'],
        'COD_FUNCT_LOC': ['!12345', '?09876'],
        'COD_NOTIF_OBJ': ['00112233', '00998877'],
        'COD_MAINT_ITEM': ['', '019283'],
        'DES_MEDIUM': ['short description', 'short description'],
        'DES_FUNCT_LOC': ['XYZ1234', 'ABC9876'],
    })
    turbines_df = pd.DataFrame({
        'COD_ELEMENT': [0],
        'TURBINE_PI_ID': ['TA00'],
        'TURBINE_LOCAL_ID': ['A0'],
        'TURBINE_SAP_COD': ['LOC000'],
        'DES_CORE_ELEMENT': ['T00'],
        'SITE': ['LOCATION'],
        'DES_CORE_PLANT': ['LOC'],
        'COD_PLANT_SAP': ['ABC'],
        'PI_COLLECTOR_SITE_NAME': ['LOC0'],
        'PI_LOCAL_SITE_NAME': ['LOC0']
    })
    return {
        'alarms': alarms_df,
        'stoppages': stoppages_df,
        'notifications': notifications_df,
        'work_orders': work_orders_df,
        'turbines': turbines_df
    }


@pytest.fixture
def pidata_dfs(base_dfs):
    pidata_df = pd.DataFrame({
        'time': [pd.Timestamp('2022-01-02 13:21:01'), pd.Timestamp('2022-03-08 13:21:01')],
        'COD_ELEMENT': [0, 0],
        'val1': [9872.0, 559.0],
        'val2': [10.0, -7.0]
    })
    return {**base_dfs, 'pidata': pidata_df}


@pytest.fixture
def scada_dfs(base_dfs):
    scada_df = pd.DataFrame({
        'TIMESTAMP': [pd.Timestamp('2022-01-02 13:21:01'), pd.Timestamp('2022-03-08 13:21:01')],
        'COD_ELEMENT': [0, 0],
        'val1': [1002.0, 56.8],
        'val2': [-98.7, 1004.2]
    })
    return {**base_dfs, 'scada': scada_df}


def create_pidata_entityset(pidata_dfs):
    return _create_entityset(pidata_dfs, es_type="pidata")


def create_scada_entityset(scada_dfs):
    return _create_entityset(scada_dfs, es_type="scada")


def test_create_pidata_missing_entities(pidata_dfs):
    error_msg = 'Missing dataframes for entities notifications.'

    pidata_dfs.pop('notifications')
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)


def test_create_scada_missing_entities(scada_dfs):
    error_msg = 'Missing dataframes for entities notifications.'

    scada_dfs.pop('notifications')
    with pytest.raises(ValueError, match=error_msg):
        create_scada_entityset(scada_dfs)


def test_create_pidata_extra_entities(pidata_dfs):
    error_msg = "Unrecognized entities extra included in dfs."

    pidata_dfs['extra'] = pd.DataFrame()
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)


def test_create_scada_extra_entities(scada_dfs):
    error_msg = "Unrecognized entities extra included in dfs."

    scada_dfs['extra'] = pd.DataFrame()
    with pytest.raises(ValueError, match=error_msg):
        create_scada_entityset(scada_dfs)


def test_missing_wo_index_columns(pidata_dfs):
    error_msg = 'Expected column "COD_ORDER" missing from notifications entity'
    pidata_dfs['notifications'].drop(columns=['COD_ORDER'], inplace=True)
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)

    error_msg = 'Expected index column "COD_ORDER" missing from work_orders entity'
    pidata_dfs['work_orders'].drop(columns=['COD_ORDER'], inplace=True)
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)


def test_wo_index_column_nonunique(pidata_dfs):
    error_msg = 'Expected index column "COD_ORDER" of work_orders entity is not unique'

    pidata_dfs['work_orders']['COD_ORDER'] = [12345, 12345]
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)


def test_missing_turbine_index_columns(pidata_dfs):
    error_msg = 'Turbines index column "COD_ELEMENT" missing from data for alarms entity'

    pidata_dfs['alarms'].drop(columns='COD_ELEMENT', inplace=True)
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)

    error_msg = 'Expected index column "COD_ELEMENT" missing from turbines entity'

    pidata_dfs['turbines'].drop(columns=['COD_ELEMENT'], inplace=True)
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)


def test_missing_time_indices(pidata_dfs):
    error_msg = 'Column "IND_LOST_GEN" associated with secondary time index ' + \
                '"DAT_END" missing from stoppages entity'
    pidata_dfs['stoppages'].drop(columns=['IND_LOST_GEN'], inplace=True)
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)

    error_msg = 'Secondary time index "DAT_END" missing from stoppages entity'
    pidata_dfs['stoppages'].drop(columns=['DAT_END'], inplace=True)
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)

    error_msg = 'Missing time index column "DAT_START" from stoppages entity'
    pidata_dfs['stoppages'].drop(columns=['DAT_START'], inplace=True)
    with pytest.raises(ValueError, match=error_msg):
        create_pidata_entityset(pidata_dfs)


def test_default_create_pidata_entityset(pidata_dfs):
    es = create_pidata_entityset(pidata_dfs)

    assert es.id == 'pidata'
    assert set(es.dataframe_dict.keys()) == set(
        ['alarms', 'turbines', 'stoppages', 'work_orders', 'notifications', 'pidata'])


def test_default_create_scada_entityset(scada_dfs):
    es = create_scada_entityset(scada_dfs)

    assert es.id == 'scada'
    assert set(es.dataframe_dict.keys()) == set(
        ['alarms', 'turbines', 'stoppages', 'work_orders', 'notifications', 'scada'])



from zephyr_ml.labeling.labeling_functions.brake_pad_presence import brake_pad_presence
from zephyr_ml.labeling.labeling_functions.converter_replacement_presence import (
    converter_replacement_presence)
from zephyr_ml.labeling.labeling_functions.planet_bearing import gearbox_replace_presence
from zephyr_ml.labeling.labeling_functions.total_power_loss import total_power_loss



from zephyr_ml.labeling.utils import denormalize


def brake_pad_presence(es, column_map={}):
    """Determines if brake pad present in stoppages.

    Args:
        es (ft.EntitySet):
            EntitySet of data to calculate total power loss across.
        column_mapping (dict):
            Optional dictionary to update default column names to the
            actual corresponding column names in the data slice. Can contain the
            following keys:
                "comments": Column that contains comments about the stoppage. Defaults
                    to "DES_COMMENTS".
                "turbine_id": Column containing the ID of the turbine associated with a
                    stoppage. Must match the index column of the 'turbines' entity.
                    Defaults to "COD_ELEMENT".
                "time_index": Column to use as the time index for the data slice.
                    Defaults to "DAT_END".

    Returns:
        label:
            Labeling function to find brake pad presence over a data slice.
        df:
            Denormalized dataframe of data to get labels from
        meta:
            Dictionary containing metadata about labeling function

    """
    comments = column_map.get('comments_column', 'DES_COMMENTS')
    turbine_id = column_map.get('turbine_id_column', 'COD_ELEMENT')
    time_index = column_map.get('time_index_column', 'DAT_END')

    def label(ds, **kwargs):
        a = ds[comments]
        a = a.fillna('')
        a = a.str.lower()
        f = any(a.apply(lambda d: ('brake' in d)
                and ('pad' in d) and ('yaw' not in d)))
        return f

    meta = {
        "target_entity_index": turbine_id,
        "time_index": time_index,
    }

    df = denormalize(es, entities=['stoppages'])

    return label, df, meta



from sigpro import SigPro


def process_signals(es, signal_dataframe_name, signal_column, transformations, aggregations,
                    window_size=None, replace_dataframe=False, **kwargs):
    '''
    Process signals using SigPro.

    Apply SigPro transformations and aggregations on the specified entity from the
    given entityset. If ``replace_dataframe=True``, then the old entity will be updated.

    Args:
        es (featuretools.EntitySet):
            Entityset to extract signals from.
        signal_dataframe_name (str):
            Name of the dataframe in the entityset containing signal data to process.
        signal_column (str):
            Name of column or containing signal values to apply signal processing pipeline to.
        transformations (list[dict]):
            List of dictionaries containing the transformation primitives.
        aggregations (list[dict]):
            List of dictionaries containing the aggregation primitives.
        window_size (str):
            Size of the window to bin the signals over. e.g. ('1h).
        replace_dataframe (bool):
            If ``True``, will replace the entire signal dataframe in the EntitySet with the
            processed signals. Defaults to ``False``, creating a new child dataframe containing
            processed signals with the suffix ``_processed``.
    '''
    signal_df = es[signal_dataframe_name]
    time_index = signal_df.ww.time_index

    for relationship in es.relationships:
        child_name = relationship.child_dataframe.ww.name
        parent_name = relationship.parent_dataframe.ww.name

        if child_name == signal_df.ww.name and parent_name == 'turbines':
            old_relationship = relationship
            groupby_index = relationship.child_column.name

    pipeline = SigPro(transformations, aggregations,
                      values_column_name=signal_column, **kwargs)

    processed_df, f_cols = pipeline.process_signal(
        signal_df,
        window=window_size,
        time_index=time_index,
        groupby_index=groupby_index,
        **kwargs
    )

    if replace_dataframe:
        es.add_dataframe(
            processed_df,
            signal_dataframe_name,
            time_index=time_index,
            index='_index')

    else:
        df_name = '{}_processed'.format(signal_df.ww.name)
        es.add_dataframe(processed_df, df_name, time_index=time_index, make_index=True,
                         index='_index')

        es.add_relationship('turbines', old_relationship.parent_column.name, df_name,
                            old_relationship.child_column.name)



"""
Preprocessing functions
"""

import sklearn.model_selection


def train_test_split(
    X,
    y,
    test_size=None,
    train_size=None,
    random_state=None,
    shuffle=True,
    stratify=None,
):
    """
    Wrapper over sklearn.model_selection.train_test_split()
    Used to split only 2 arrays at once: X (features) and y (labels)

    Split arrays or matrices into random train and test subsets.
    """
    return sklearn.model_selection.train_test_split(
        X, y, test_size, train_size, random_state, shuffle, stratify
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

try:
    with open('README.md') as readme_file:
        readme = readme_file.read()
except IOError:
    readme = ''

try:
    with open('HISTORY.md') as history_file:
        history = history_file.read()
except IOError:
    history = ''

install_requires = [
    'numpy>=1.19.5,<1.27.0',
    'pandas>=1,<2',
    'composeml>=0.1.6,<0.10',
    'featuretools>=1.0.0,<2.0.0',
    'mlblocks>=0.6.0,<0.7',
    'sigpro>=0.2.0',
    'xgboost>=0.72.1,<1',
    'jupyter==1.0.0',
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
    'jupyter>=1.0.0,<2',
    'rundoc>=0.4.3,<0.5',
    'invoke',
]

development_requires = [
    # general
    'bumpversion>=0.5.3,<0.6',
    'pip>=9.0.1',
    'watchdog>=0.8.3,<0.11',

    # docs
    'docutils>=0.12,<0.18',
    'lxml_html_clean>=0.2.2',
    'm2r2>=0.2.5,<0.3',
    'nbsphinx>=0.5.0,<0.7',
    'Sphinx>=3,<3.3',
    'sphinx_rtd_theme>=0.2.4,<0.5',
    'autodocsumm>=0.1.10',
    'mistune>=0.7,<2.0',
    'Jinja2>=2,<3.1',

    # fails on Sphinx < v3.4
    'alabaster<=0.7.12',
    # fails on Sphins < v5.0
    'sphinxcontrib-applehelp<1.0.8',
    'sphinxcontrib-devhelp<1.0.6',
    'sphinxcontrib-htmlhelp<2.0.5',
    'sphinxcontrib-serializinghtml<1.1.10',
    'sphinxcontrib-qthelp<1.0.7',

    # style check
    'flake8>=3.7.7,<4',
    'isort>=4.3.4,<5',

    # fix style issues
    'autoflake>=1.1,<2',
    'autopep8>=1.4.3,<2',

    # distribute on PyPI
    'twine>=1.10.0,<4',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1,<6',
    'tox>=2.9.1,<4',

    # Jupyter
    'jupyter>=1.0.0',
]

setup(
    author='MIT Data To AI Lab',
    author_email='dai-lab@mit.edu',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    description='Prediction engineering methods for Draco.',
    entry_points={
        'mlblocks': [
            'primitives=zephyr_ml:MLBLOCKS_PRIMITIVES',
            'pipelines=zephyr_ml:MLBLOCKS_PIPELINES'
        ],
    },
    extras_require={
        'test': tests_require,
        'dev': development_requires + tests_require,
    },
    install_package_data=True,
    install_requires=install_requires,
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='zephyr Draco Prediction Engineering',
    name='zephyr-ml',
    packages=find_packages(include=['zephyr_ml', 'zephyr_ml.*']),
    python_requires='>=3.8,<3.13',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/sintel-dev/zephyr',
    version='0.0.5.dev1',
    zip_safe=False,
)



"""
Postprocessing functions.
"""

import logging

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import sklearn
from sklearn import metrics

LOGGER = logging.getLogger(__name__)

METRICS = {
    "accuracy": sklearn.metrics.accuracy_score,
    "precision": sklearn.metrics.precision_score,
    "recall": sklearn.metrics.recall_score,
    "f1": sklearn.metrics.f1_score,
}


class FindThreshold:
    """Find Optimal Threshold.

    This class find the optimal threshold value that produces
    the highest metric score. In the fit phase, it detects
    the best threshold based on the given metric. In the
    produce phase, it applies the found threshold on the
    predicted values.

    This is intended for classification problems.

    Args:
        metric (str):
            String representing which metric to use.
    """

    def __init__(self, metric="f1"):
        self._metric = "f1"
        self._threshold = None

    def fit(self, y_true, y_proba):
        """Find the threshold that obtains the best metric value.

        Args:
            y_true (Series or ndarray):
                ``pandas.Series`` or ``numpy.ndarray`` ground truth target values.
            y_proba (Series or ndarray):
                ``pandas.Series`` or ``numpy.ndarray`` predicted target values' probabilities.
        """
        if y_proba.ndim > 1:
            y_proba = y_proba[:, 1]

        RANGE = np.arange(0, 1, 0.01)

        scores = list()
        scorer = METRICS[self._metric]
        for thresh in RANGE:
            y = [1 if x else 0 for x in y_proba > thresh]
            scores.append(scorer(y_true, y))

        threshold = RANGE[np.argmax(scores)]
        LOGGER.info(f"best threshold found at {threshold}")

        self._threshold = threshold
        self._scores = scores

    def apply_threshold(self, y_proba):
        """Apply threshold on predicted values.

        Args:
            y_pred (Series):
                ``pandas.Series`` predicted target values' probabilities.

        Return:
            tuple:
                * list of predicted target valeus in binary codes.
                * detected float value for threshold.
                * list of scores obtained at each threshold.
        """
        if y_proba.ndim > 1:
            y_proba = y_proba[:, 1]

        binary = [1 if x else 0 for x in y_proba > self._threshold]
        return binary, self._threshold, self._scores


def confusion_matrix(
        y_true,
        y_pred,
        labels=None,
        sample_weight=None,
        normalize=None):
    conf_matrix = metrics.confusion_matrix(
        y_true, y_pred, labels=labels, sample_weight=sample_weight, normalize=normalize
    )
    fig = plt.figure()
    ax = fig.add_axes(sns.heatmap(conf_matrix, annot=True, cmap="Blues"))

    ax.set_title("Confusion Matrix\n")
    ax.set_xlabel("\nPredicted Values")
    ax.set_ylabel("Actual Values")

    ax.xaxis.set_ticklabels(["False", "True"])
    ax.yaxis.set_ticklabels(["False", "True"])

    return conf_matrix, fig


def roc_auc_score_and_curve(
    y_true, y_proba, pos_label=None, sample_weight=None, drop_intermediate=True
):
    if y_proba.ndim > 1:
        y_proba = y_proba[:, 1]
    fpr, tpr, _ = metrics.roc_curve(
        y_true,
        y_proba,
        pos_label=pos_label,
        sample_weight=sample_weight,
        drop_intermediate=drop_intermediate,
    )
    ns_probs = [0 for _ in range(len(y_true))]
    ns_fpr, ns_tpr, _ = metrics.roc_curve(
        y_true,
        ns_probs,
        pos_label=pos_label,
        sample_weight=sample_weight,
        drop_intermediate=drop_intermediate,
    )

    auc = metrics.roc_auc_score(y_true, y_proba)
    fig, ax = plt.subplots(1, 1)

    ax.plot(fpr, tpr, "ro")
    ax.plot(fpr, tpr)
    ax.plot(ns_fpr, ns_tpr, linestyle="--", color="green")

    ax.set_ylabel("True Positive Rate")
    ax.set_xlabel("False Positive Rate")
    ax.set_title("AUC: %.3f" % auc)

    return auc, fig



from itertools import chain

import featuretools as ft

from zephyr_ml.metadata import get_mapped_kwargs


def _validate_data(dfs, es_type, es_kwargs):
    """Validate data by checking for required columns in each entity"""
    if not isinstance(es_type, list):
        es_type = [es_type]

    entities = set(
        chain(
            [
                "alarms",
                "stoppages",
                "work_orders",
                "notifications",
                "turbines",
                *es_type,
            ]
        )
    )

    if set(dfs.keys()) != entities:
        missing = entities.difference(set(dfs.keys()))
        extra = set(dfs.keys()).difference(entities)
        msg = []
        if missing:
            msg.append("Missing dataframes for entities {}.".format(
                ", ".join(missing)))
        if extra:
            msg.append(
                "Unrecognized entities {} included in dfs.".format(
                    ", ".join(extra))
            )

        raise ValueError(" ".join(msg))

    turbines_index = es_kwargs["turbines"]["index"]
    work_orders_index = es_kwargs["work_orders"]["index"]

    if work_orders_index not in dfs["work_orders"].columns:
        raise ValueError(
            'Expected index column "{}" missing from work_orders entity'.format(
                work_orders_index
            )
        )

    if work_orders_index not in dfs["notifications"].columns:
        raise ValueError(
            'Expected column "{}" missing from notifications entity'.format(
                work_orders_index
            )
        )

    if not dfs["work_orders"][work_orders_index].is_unique:
        raise ValueError(
            'Expected index column "{}" of work_orders entity is not '
            "unique".format(work_orders_index)
        )

    if turbines_index not in dfs["turbines"].columns:
        raise ValueError(
            'Expected index column "{}" missing from turbines entity'.format(
                turbines_index
            )
        )

    if not dfs["turbines"][turbines_index].is_unique:
        raise ValueError(
            'Expected index column "{}" of turbines entity is not unique.'.format(
                turbines_index
            )
        )

    for entity, df in dfs.items():
        if turbines_index not in df.columns:
            raise ValueError(
                'Turbines index column "{}" missing from data for {} entity'.format(
                    turbines_index, entity
                )
            )

        time_index = es_kwargs[entity].get("time_index", False)
        if time_index and time_index not in df.columns:
            raise ValueError(
                'Missing time index column "{}" from {} entity'.format(
                    time_index, entity
                )
            )

        secondary_time_indices = es_kwargs[entity].get(
            "secondary_time_index", {})
        for time_index, cols in secondary_time_indices.items():
            if time_index not in df.columns:
                raise ValueError(
                    'Secondary time index "{}" missing from {} entity'.format(
                        time_index, entity
                    )
                )
            for col in cols:
                if col not in df.columns:
                    raise ValueError(
                        (
                            'Column "{}" associated with secondary time index "{}" '
                            "missing from {} entity"
                        ).format(col, time_index, entity)
                    )


def validate_scada_data(dfs, new_kwargs_mapping=None):
    """
    SCADA data is signal data from the Original Equipment Manufacturer Supervisory Control
    And Data Acquisition (OEM-SCADA) system, a signal data source.
    """
    entity_kwargs = get_mapped_kwargs("scada", new_kwargs_mapping)
    _validate_data(dfs, "scada", entity_kwargs)
    return entity_kwargs


def validate_pidata_data(dfs, new_kwargs_mapping=None):
    """
    PI data is signal data from the operator's historical Plant Information (PI) system.
    """
    entity_kwargs = get_mapped_kwargs("pidata", new_kwargs_mapping)
    _validate_data(dfs, "pidata", entity_kwargs)
    return entity_kwargs


def validate_vibrations_data(dfs, new_kwargs_mapping=None):
    """
    Vibrations data is vibrations data collected on Planetary gearboxes in turbines.
    """
    entities = ["vibrations"]

    pidata_kwargs, scada_kwargs = {}, {}
    if "pidata" in dfs:
        pidata_kwargs = get_mapped_kwargs("pidata", new_kwargs_mapping)
        entities.append("pidata")
    if "scada" in dfs:
        scada_kwargs = get_mapped_kwargs("scada", new_kwargs_mapping)
        entities.append("scada")

    entity_kwargs = {
        **pidata_kwargs,
        **scada_kwargs,
        **get_mapped_kwargs("vibrations", new_kwargs_mapping),
    }
    _validate_data(dfs, entities, entity_kwargs)
    return entity_kwargs


VALIDATE_DATA_FUNCTIONS = {
    "scada": validate_scada_data,
    "pidata": validate_pidata_data,
    "vibrations": validate_vibrations_data,
}


def _create_entityset(entities, es_type, new_kwargs_mapping=None):

    validate_func = VALIDATE_DATA_FUNCTIONS[es_type]
    es_kwargs = validate_func(entities, new_kwargs_mapping)

    # filter out stated logical types for missing columns
    for entity, df in entities.items():
        es_kwargs[entity]["logical_types"] = {
            col: t
            for col, t in es_kwargs[entity]["logical_types"].items()
            if col in df.columns
        }

    turbines_index = es_kwargs["turbines"]["index"]
    work_orders_index = es_kwargs["work_orders"]["index"]

    relationships = [
        ("turbines", turbines_index, "alarms", turbines_index),
        ("turbines", turbines_index, "stoppages", turbines_index),
        ("turbines", turbines_index, "work_orders", turbines_index),
        ("turbines", turbines_index, es_type, turbines_index),
        ("work_orders", work_orders_index, "notifications", work_orders_index),
    ]

    es = ft.EntitySet()
    es.id = es_type

    for name, df in entities.items():
        es.add_dataframe(dataframe_name=name, dataframe=df, **es_kwargs[name])

    for relationship in relationships:
        parent_df, parent_column, child_df, child_column = relationship
        es.add_relationship(parent_df, parent_column, child_df, child_column)

    return es
